/*=======================================================
  Author:   R. Muth
  Version:  $Id: aliasing.c,v 1.4 1998/11/18 19:40:50 muth Exp muth $
  Comment:  This file contains code for the alias analysis
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

/*=======================================================
  =======================================================*/
#undef __PROC__
#define __PROC__ "DoMemoryAccessesConflict"

extern NUMBER DoMemoryAccessesConflict(INDEX ains, INDEX bins)
{
    
    REG    areg = AINS_REG_B(ains);
    INT64  alo = AINS_DATA(ains);
    INT64  ahi;

    REG    breg = AINS_REG_B(bins);
    INT64  blo = AINS_DATA(bins);
    INT64  bhi;
    BOOL   verbose = 0;

    CASSERT( AINS_IS_LOAD(ains) || AINS_IS_STORE(ains),("load/store ins expected\n"));

    if( verbose )
    {
        printf("\n");
        
        printf("ins: %5d off: %4x loaddef: %5d loadoff: %4x\n",ains,AINS_DATA(ains),AINS_LOAD_DEF(ains),AINS_LOAD_OFF(ains));
        printf("ins: %5d off: %4x loaddef: %5d loadoff: %4x\n",bins,AINS_DATA(bins),AINS_LOAD_DEF(bins),AINS_LOAD_OFF(bins));
    }
    
    
    if( (AINS_IS_LOAD(ains) && AINS_REG_C(ains) == REG_ZERO) || (AINS_IS_LOAD(bins) && AINS_REG_C(bins) == REG_ZERO) )
    {
        if( verbose )
            printf("-> no \n");
        return CONFLICT_NO;        
    }
    
          
    if( (areg == REG_SP && (AINS_FLAGS(bins) & FI_EVAL_B)) ||
        (breg == REG_SP && (AINS_FLAGS(ains) & FI_EVAL_B)) )
    {
        if( verbose )
            printf("-> no \n");
        return CONFLICT_NO;        
    }

    /* if one is a stack pointer and the other is not and they are not realated */
    
       
    if( AINS_LOAD_DEF(ains) != AINS_LOAD_DEF(bins) &&
        ((areg == REG_SP && breg != REG_SP) || (breg == REG_SP && areg != REG_SP))  )
        
    {
        INDEX afun = ABBL_FUN(AINS_BBL(ains));
        INDEX bfun = ABBL_FUN(AINS_BBL(bins));
        
        if( afun == bfun && !AFUN_HAS_STK_ALIAS(afun) ) /* this include bad aliases */
        {
            if( verbose )
                printf("-> no \n");
            return CONFLICT_NO;
        }
    }

#if 0
    if( AINS_BBL(ains) != AINS_BBL(bins) )
    {
        if( verbose )
            printf("-> maybe \n");
        return CONFLICT_MAYBE;
    }
#endif
    
    if( (AINS_FLAGS(bins) & FI_EVAL_B) &&  (AINS_FLAGS(ains) & FI_EVAL_B) )
    {
        alo += AINS_B_EVAL(ains);
        blo += AINS_B_EVAL(bins);
    }
    else
    {
        if( AINS_LOAD_DEF(ains) != AINS_LOAD_DEF(bins) )
            return CONFLICT_MAYBE;
        alo += AINS_LOAD_OFF(ains);
        blo += AINS_LOAD_OFF(bins);
    }
    
    ahi = AinsMemWidth(ains) + alo;
    if( AINS_TYPE(ains) == IT_UST || AINS_TYPE(ains) == IT_ULD )
    {
        ahi += 7;
        alo -= 7;
    }
    
    bhi = AinsMemWidth(bins) + blo;
    if( AINS_TYPE(bins) == IT_UST || AINS_TYPE(bins) == IT_ULD )
    {
        bhi += 7;
        blo -= 7;
    }

    if( ahi <= blo )
    {
        if( verbose )
            printf("-> no \n");
        return CONFLICT_NO;
    }
    
    if( bhi <= alo )
    {
        if( verbose )
            printf("-> no \n");
        return CONFLICT_NO;
    }
    
    if( alo == blo )
    {
        if( verbose )
            printf("-> exact \n");
        return CONFLICT_EXACT;
    }
    
    
    if( verbose )
        printf("-> match \n");
    return
        CONFLICT_MATCH;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsFindDefInExtBbl"

static INDEX AinsFindDefInExtBbl(INDEX iins, REG reg)
{
    INDEX ibbl = AINS_BBL(iins);

    iins = AINS_PREV(iins);
    
    for(;;)
    {
        for( ; VALID(iins); iins = AINS_PREV(iins) )    
        {
            if( reg == AINS_REG_C(iins) ) return iins;
        }
        if( AbblNpreds(ibbl) != 1 ) return INVALID;

        ibbl = AEDG_HEAD(ABBL_PRED(ibbl));
        iins = ABBL_INS_LAST(ibbl);
    }
}


/*=======================================================
@COM
This routine works with extended basic blocks so it is
important to run the Dead Code Remover first

Requires DeadCode removal
Requires Stack Usage Info
@END
=======================================================*/

extern void ComputeAliasAuxiliaryInformation(FILE *fp)
{
    INDEX ibbl;
    INDEX iins;
    INDEX iins_def;
    NUMBER offset_def;
    REG reg_def;
    clock_t time = clock();
  
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            AINS_LOAD_DEF(iins) = INVALID;
            AINS_LOAD_OFF(iins) = INVALID;
                    
            switch( AINS_TYPE(iins) )
            {
            case IT_ILD:
            case IT_FLD:
            case IT_ULD:
            case IT_LLD:
            case IT_UST:
            case IT_FST:
            case IT_IST:
            case IT_CST:
                if( AINS_FLAGS(iins) & FI_EVAL_B )
                    break;
                
                offset_def = 0;
                reg_def = AINS_REG_B(iins);
                iins_def = iins;
                for(;;)
                {
                    iins_def = AinsFindDefInExtBbl(iins_def,reg_def);
                    if( !VALID(iins_def) )
                    {
                        /* this uniquely describes a bbl and a register and  cannot be confused with an instruction*/
                        iins_def = CONFIG_NINS + AINS_BBL(iins_def) * 64 + reg_def;   
                        break;
                    }
                    else if( AINS_CODE(iins_def) == I_lda )
                    {
                        offset_def += AINS_DATA(iins_def);
                        reg_def = AINS_REG_B(iins_def);
                    }
                    else
                    {
                        break;
                    }
                }

                AINS_LOAD_DEF(iins) = iins_def;
                AINS_LOAD_OFF(iins) = offset_def;

                break;
            default:
                break;
            }
        }
        
    }
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[ComAliasing] {%d ms}\n",time);
}

/*=======================================================
=======================================================*/

