/*=======================================================
  Author:   R. Muth
  Version:  $Id: dominator.c,v 1.19 1998/10/29 18:11:06 muth Exp muth $
  Comment:  This file contains the code for the dominator and loop computation
=======================================================*/

#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"


#define ABBL_DFS_NUM(i)               ABBL_U1(i).doubleword.d1
#define ABBL_DFS_PARENT(i)            ABBL_U1(i).doubleword.d2
#define ABBL_DOM_TREE_FIRST(i)        ABBL_U2(i).doubleword.d1
#define ABBL_DOM_TREE_NEXT(i)         ABBL_U2(i).doubleword.d2



static INDEX AbblNCA(INDEX ibbl,INDEX jbbl)
{
    for(;;)
    {
        while(ABBL_DFS_NUM(ibbl) > ABBL_DFS_NUM(jbbl))
            ibbl = ABBL_PRE_DOM(ibbl);
        while(ABBL_DFS_NUM(ibbl) < ABBL_DFS_NUM(jbbl))
            jbbl = ABBL_PRE_DOM(jbbl);
        if( ibbl == jbbl )
            return ibbl;
    }
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblPreDomDfs"

static INDEX AbblPreDomDfs(INDEX ibbl,INDEX no)
{
    INDEX iedg;
    
    CFATAL( VALID(ABBL_DFS_NUM(ibbl)),("bbl %d already has dfs no\n",ibbl));

    ABBL_DFS_NUM(ibbl) = no;
    no += 1;

    if( ABBL_IS_LAST(ibbl) ) return no;
    
    FOREACH_BBL_SUCC(ibbl,iedg)
    {
        INDEX succ;
        
        if( AEDG_IS_INTER(iedg) )
        {
#if 1
            if( AEDG_TYPE(iedg)!=ET_FUN_CALL && AEDG_TYPE(iedg)!=ET_FUN_RETURN )
            {
                INDEX ifun = ABBL_FUN(ibbl);
                ERROR(("unexpected inter edge in %d [%s]\n",ifun,AFUN_NAME(ifun)));
                PrintEdg(stdout,iedg);

                /* eventhough, there might be hell edges coming into this functions
                   init node they should never be processed
                   */
            }
#endif
            continue;
        }
        
        succ = AEDG_TAIL(iedg);
        if( !VALID(ABBL_DFS_NUM(succ)) )
        {
            ABBL_PRE_DOM(succ) = ibbl;
            ABBL_DFS_PARENT(succ) = ibbl;
            no = AbblPreDomDfs(succ,no);
        }
    }
    
    return no;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPreDomIter"

extern BOOL AfunPreDomIter(INDEX ifun)
{
    INDEX ibbl;
    BOOL change = 0;

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;
        FOREACH_BBL_SUCC(ibbl,iedg)
        {
            INDEX hd,tl;
            
            if( AEDG_IS_INTER(iedg) )
            {
#if 0                
                CASSERT(AEDG_TYPE(iedg) == ET_FUN_CALL || AEDG_TYPE(iedg) ==  ET_FUN_RETURN,
                        ("unexpected interprocedural edge %d\n",iedg));
#endif                
                continue;
            }

            hd = AEDG_HEAD(iedg);
            tl = AEDG_TAIL(iedg);

            if( hd == tl ) continue;
            
            hd = AbblNCA(hd,tl);
            if( ABBL_DFS_NUM(hd) < ABBL_DFS_NUM(ABBL_PRE_DOM(tl)))
            {
                ABBL_PRE_DOM(tl) = hd;
                change = 1;
            }
            
        }
    }
    
    return change;
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPreDomCompute"

static void AfunPreDomCompute(INDEX ifun)
{
    INDEX ibbl;
    NUMBER no=0;

    CERROR( !AfunIsLoopAnalyzable(ifun),("weird\n"));

    FOREACH_FUN_BBL(ifun, ibbl)
    {
        ABBL_DFS_PARENT(ibbl) = INVALID;
        ABBL_DFS_NUM(ibbl) = INVALID;
        ABBL_PRE_DOM(ibbl) = INVALID;
        no++;
    }

    if( no != AbblPreDomDfs(AFUN_BBL_FIRST(ifun),0) )
    {
        ERROR(("weird\n"));
    }
    
    while( AfunPreDomIter(ifun) )
        ;
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblPostDomDfs"

static INDEX AbblPostDomDfs(INDEX ibbl,INDEX no)
{
    INDEX iedg;
    
    CFATAL( VALID(ABBL_DFS_NUM(ibbl)),("bbl %d already has dfs no\n",ibbl));

    ABBL_DFS_NUM(ibbl) = no;
    no += 1;

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        INDEX pred;
        
        if( AEDG_IS_INTER(iedg) )
        {
            if( AEDG_TYPE(iedg) != ET_FUN_CALL && AEDG_TYPE(iedg) !=  ET_FUN_RETURN)
            {
                    ERROR(("Unexpected interprocedural edge\n"));                
                    PrintEdg(stdout,iedg);
            }
    
            continue;
        }
        
        pred = AEDG_HEAD(iedg);
        if( !VALID(ABBL_DFS_NUM(pred)) )
        {
            ABBL_POST_DOM(pred) = ibbl;
            no = AbblPostDomDfs(pred,no);
        }
    }
    
    return no;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPostDomIter"

static BOOL AfunPostDomIter(INDEX ifun)
{
    INDEX ibbl;
    BOOL change =0;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX iedg;
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            INDEX hd,tl;
            
            if( AEDG_IS_INTER(iedg) )
            {
                CASSERT(AEDG_TYPE(iedg) == ET_FUN_CALL || AEDG_TYPE(iedg) == ET_FUN_RETURN,
                        ("unexpected interprocedural edge %d\n",iedg));
                continue;
            }

            tl = AEDG_HEAD(iedg);   /* reversed!!! */
            hd = AEDG_TAIL(iedg);   /* reversed!!! */

            if( hd == tl ) continue;
                
            if( ABBL_DFS_NUM(hd) > ABBL_DFS_NUM(tl) )
            {
                do
                    hd = ABBL_POST_DOM(hd);
                while( ABBL_DFS_NUM(hd) > ABBL_DFS_NUM(tl) );
            }
            else
            {
                do
                    tl = ABBL_POST_DOM(tl);
                while( ABBL_DFS_NUM(hd) < ABBL_DFS_NUM(tl) );
                
                hd = tl;
                tl = AEDG_HEAD(iedg);
            }
      
            if( ABBL_DFS_NUM(hd) < ABBL_DFS_NUM(ABBL_POST_DOM(tl)))
            {
                ABBL_POST_DOM(tl) = hd;
                change = 1;
            }
        }
    }
    
    return change;
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunPostDomCompute"

extern BOOL AfunPostDomCompute(INDEX ifun)
{
    INDEX ibbl;
    NUMBER no=0;

    if( !AfunIsLoopAnalyzable(ifun) )
        return 0;
        
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_DFS_PARENT(ibbl) = INVALID;
        ABBL_DFS_NUM(ibbl) = INVALID;
        ABBL_POST_DOM(ibbl) = INVALID;
        no++;
    }

    if( no != AbblPostDomDfs(AFUN_BBL_LAST(ifun),0) )
        return 0;
    
    while( AfunPostDomIter(ifun) )
        ;

    return 1;
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "Dominates"


extern BOOL Dominates(INDEX bbl1, INDEX bbl2)
{
    while( VALID(bbl2) )
    {
        if( bbl1 == bbl2 ) return TRUE;
        bbl2 = ABBL_PRE_DOM(bbl2);
    }

    return FALSE;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IsBackEdge"

extern BOOL IsBackEdge(INDEX edge)
{
   return Dominates(AEDG_TO(edge), AEDG_FROM(edge));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblAddToLoop"

static void AbblAddToLoop(INDEX ibbl,INDEX loop)
{

#ifdef DEBUG_LOOP
        printf("insert %d into %d loop\n",ibbl,loop);
#endif

    CASSERT( ABBL_FUN(ibbl) == ABBL_FUN(loop),("bbl and loop are in different funs"));
    CASSERT( ABBL_IS_LOOP_HEADER(loop),("loop is not marked as loop headers\n"));
    CASSERT( loop != ibbl,("loop and bbl are identical\n"));
    
    CERROR( VALID(ABBL_LOOP_HEADER(ibbl)),("bbl is already in some loop\n"));
    CERROR( VALID(ABBL_LOOP_NEXT_MEMBER(ibbl)),("bbl has already a brother\n"));
    
    ABBL_LOOP_HEADER(ibbl) = loop;
    ABBL_LOOP_NEXT_MEMBER(ibbl) = ABBL_LOOP_FIRST_MEMBER(loop);
    ABBL_LOOP_FIRST_MEMBER(loop)= ibbl;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblLoopInsert"

static void AbblLoopInsert(INDEX ibbl, INDEX loop)
{
   INDEX iedg;

   /*
     we want to insert ibbl to the loop if it already is in some other loop we just
     continue with its header
     */
   
#ifdef DEBUG_LOOP      
   printf("LoopInsert %d into %d loop\n",ibbl,loop);
#endif
   
   if( ibbl == loop ) return;
   
   
   if( !VALID( ABBL_LOOP_HEADER(ibbl) ) )
   {
       AbblAddToLoop(ibbl,loop);
       FOREACH_BBL_PRED(ibbl,iedg)
       {
           if(AEDG_IS_INTER(iedg)) /* skip inter-procedural edges */
               continue;
           
           AbblLoopInsert(AEDG_FROM(iedg),loop);
       }
   }
   else
   {
       AbblLoopInsert(ABBL_LOOP_HEADER(ibbl),loop);
   }
   
}

/*=======================================================
Recursivle travers the Dominator tree in a postoerder fashion.
  =======================================================*/
#undef __PROC__
#define __PROC__ "AbblDoHeader"


static void AbblDoHeader(INDEX ibbl)
{
    INDEX iedg;

#ifdef DEBUG_LOOP
    printf("DoHeader: %d\n",ibbl);
#endif
    
    if( VALID(ABBL_DOM_TREE_FIRST(ibbl)) )
        AbblDoHeader(ABBL_DOM_TREE_FIRST(ibbl));

    if( VALID(ABBL_DOM_TREE_NEXT(ibbl)) )
        AbblDoHeader(ABBL_DOM_TREE_NEXT(ibbl));

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        if( AEDG_IS_INTER(iedg) )
        {
            if( AEDG_TYPE(iedg) != ET_FUN_RETURN && AEDG_TYPE(iedg) != ET_FUN_CALL && !ABBL_IS_FIRST(ibbl) )
            {
                /* occasionally there might be a warning here if there is a hell edge coming into
                    the init bbl */
                ERROR(("bad edge type\n"));
                PrintEdg(stdout,iedg);
            }
            continue;
        }
            
        if( !IsBackEdge(iedg) ) continue;
#ifdef DEBUG_LOOP
        printf("doing loop header %d (edge %d)\n",ibbl,iedg);
#endif
        
        ABBL_FLAGS(ibbl) |= FB_LOOP_HEADER;
        AbblLoopInsert(AEDG_FROM(iedg),ibbl);
    }
    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunComputeLoops"

static void AfunComputeLoops(INDEX ifun)
{
    INDEX ibbl;
    INDEX root;
    
    /* we process the backedge first whose loopheader is farther down in
       the dominator tree, ie. close to the leaves.
       We achieve this by first creating a topological order of loop headers
       */


    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_FLAGS(ibbl) &= ~(FB_LOOP_HEADER);
        ABBL_LOOP_HEADER(ibbl) = INVALID;
        ABBL_LOOP_NEXT_MEMBER(ibbl) = INVALID;
        ABBL_LOOP_FIRST_MEMBER(ibbl) = INVALID;
        ABBL_DOM_TREE_NEXT(ibbl) = INVALID;
        ABBL_DOM_TREE_FIRST(ibbl) = INVALID;
    }

    root = INVALID;
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX parent = ABBL_PRE_DOM( ibbl );
 
        /* If bbl doesn't have a dominator it is a top node
           in its dominator tree.  Nothings needs to be done. */
        
        if ( !VALID(parent) )
        {
            root = ibbl;
            continue;
        }

        ABBL_DOM_TREE_NEXT(ibbl) = ABBL_DOM_TREE_FIRST(parent);
        ABBL_DOM_TREE_FIRST(parent)= ibbl;
    }

#ifdef DEBUG_LOOP
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        printf("node: %d dom: %d  1st: %d  next: %d\n",ibbl,ABBL_PRE_DOM(ibbl), ABBL_DOM_TREE_FIRST(ibbl),ABBL_DOM_TREE_NEXT(ibbl)); 
    }
#endif
    
    CASSERT( VALID(root),("weird\n"));
    CASSERT( ABBL_IS_FIRST(root),("weird\n"));

    AbblDoHeader(root);
}

/*=======================================================
@COM
Computes the immediate dominator for bbls in suitable functions.
Uses dominators to compute natural loop (Dragon book p. 603).
Some functions, eg. with escaping edges are not suitable.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComDomsAndLoops"

extern void ComDomsAndLoops(FILE *fp) 
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Domloop");
    INDEX ifun;
    
    NUMBER count = 0;
    NUMBER total = 0;
    clock_t time = clock();
    

   /* for each function compute loop info */
       
   FOREACH_LIVE_FUN(ifun)
   {
          total++;
          if( !AfunIsLoopAnalyzable(ifun) ) continue;
             
          if( verbose )
              printf("\t[*] DOM %d [%s]\n",ifun, AFUN_NAME(ifun));

          AfunPreDomCompute(ifun);

          if( verbose )
              printf("\t[*] LOOP %d [%s]\n",ifun, AFUN_NAME(ifun));

          
          AfunComputeLoops(ifun);

          if( verbose )
              printf("\t[*] DONE %d [%s]\n",ifun, AFUN_NAME(ifun));

          count++;
          
   }
   
   time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
   InfoStatus(fp,"\t[ComDomloop] Analyzed funs: %d/%d {%d ms}\n",count,total,time);
}
/*=======================================================
=======================================================*/






