/*=======================================================
  Author:   R. Muth
  Version:  $Id: emit.c,v 1.30 1998/11/18 22:36:00 saw Exp muth $
  Comment:  This file contains the high level code generation routines
            handling the binary
            This code abuses the eval field to store the new
            address of a basic block
=======================================================*/

#include "global.h"
#include "functions.h"
#include "flags.h"
#include "opcode.h"
#include "system.coff.h"

#define ABBL_NEW_ADDR(i) ABBL_U3(i).doubleword.d1

/*=======================================================
@FUN
Mark certain basic blocks to be aligned (16 byte boundary).
the scheduler should take this information into account
but currently doesn't.
Rephrasing this: After scheduling enough noops should have been
added so that at least the marked bbls are aligned.
Currently enforce the alignment when assembling the code, though.

The order of stuff in here is extremely important!!!
Don't mess with it unless you know what you are doing.
Always do some experiments afterwards to see whether
you changes are any good.
@END
=======================================================*/


#undef __PROC__
#define __PROC__ "EmitComputeAlignment"

extern void EmitComputeAlignment(FILE *fp,INDEX xfun)
{
    INDEX  ibbl;
    NUMBER acount=0;
    NUMBER tcount=0;

    FOREACH_FUN_BBL(xfun,ibbl)
    {
        INDEX prev = ABBL_PREV(ibbl);
#if 0
        INDEX next = ABBL_NEXT(ibbl);
        INDEX n    = AbblNInss(ibbl);
#endif
        tcount++;

        ABBL_FLAGS(ibbl) &= ~(FB_ALIGNED_4|FB_ALIGNED_2);


        if( !VALID(prev) )
        {
            continue;
        }
                
        if( ABBL_TYPE(prev) == BT_EXTRA ||
            ABBL_TYPE(prev) == BT_RET ||
            ABBL_TYPE(prev) == BT_JUMP ||
            ABBL_TYPE(prev) == BT_SWITCH ||
            ABBL_TYPE(prev) == BT_UNCOND ||
            /* ABBL_TYPE(prev) == BT_CALL || */ /* this only makes sense if noops are scheduled in the calling bbls */
            ABBL_TYPE(prev) == BT_HALT )
        {
            ABBL_FLAGS(ibbl) |= FB_ALIGNED_4;
            acount++;   
            continue;
        }

#if 0                
        if( (ABBL_TYPE(ibbl) == BT_CALL ||
             ABBL_TYPE(ibbl) == BT_EXTRA ||
             ABBL_TYPE(ibbl) == BT_RET ||
             ABBL_TYPE(ibbl) == BT_JUMP ||
             ABBL_TYPE(ibbl) == BT_SWITCH || 
             ABBL_TYPE(ibbl) == BT_HALT ) && n <= 3 )
        {
            continue;
        }
#endif
            
#if 1
        if( ABBL_TYPE(ibbl) == BT_BRANCH )
        {
            if( AEDG_TAIL(ABBL_EDG_FALSE(ibbl)) == ibbl ||
                AEDG_TAIL(ABBL_EDG_TRUE(ibbl)) == ibbl )
            {
                
                ABBL_FLAGS(ibbl) |= FB_ALIGNED_4;
                acount++;
                continue;
            }
            
        }
#endif    

#if 0
        if( ABBL_IS_LOOP_HEADER(ibbl) )
        {
            ABBL_FLAGS(ibbl) |= FB_ALIGNED_4;
            acount++;
            continue;
        }
#endif
        

#if 1
        if( ABBL_IS_HOT(ibbl) && !ABBL_IS_HOT(prev) )
        {
            ABBL_FLAGS(ibbl) |= FB_ALIGNED_4;
            acount++;
#if 0
            printf("aligning: %d  prev: %d\n",ibbl,prev);
            PrintBblSuper(stdout,ibbl);
            PrintBblSuper(stdout,prev);
#endif
            continue;
            
        }
#endif

#if 0
        if( ABBL_IS_HOT(ibbl) && (n & 1 == 0) )
        {
            ABBL_FLAGS(ibbl) |= FB_ALIGNED_2;
            acount++;
            continue;
        }
#endif        
    }

    InfoStatus(fp,"\t[EmitAlign] Aligned Bbls: %d/%d\n",acount,tcount);
}


/*=======================================================
=======================================================*/
static INDEX Align(NUMBER align,INDEX isec, INDEX offset, INDEX noop)
{
    CASSERT( align >= 0,("weird\n"));
    
    while( offset & (align-1) )
    {
        AsecWriteData32(isec,offset,Assemble(&AINS(noop)));
        offset++;
    }
    return offset;
}
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "EmitGenerateCode"

extern void EmitGenerateCode(FILE *fp,INDEX xfun)
{
    INDEX ft_edge = INVALID;
    INDEX unop = AinsNew();
    INDEX ibbl;
    INDEX iins;
    INDEX isec = SysGetTextSection();
    NUMBER offset = 0;
    NUMBER count=0;
    INT64  weight=0;
    

    AinsMakeUnop(unop);

    FOREACH_FUN_BBL(xfun,ibbl)
    {
        /*
        offset  = Align(ABBL_ALIGN(ibbl), isec, offset, unop);
        */
        
        if( VALID(ft_edge) )
        {
            if(  AEDG_TAIL(ft_edge) == ibbl )
                ft_edge = INVALID;
            else if( ABBL_HAS_INSS(ibbl) )
            {
                ERROR(("Could not resolve fall thru from bbl %d to bbl %d\n",
                       AEDG_HEAD(ft_edge),AEDG_TAIL(ft_edge)));
                PrintEdgSuper(stdout,ft_edge);
                ft_edge = INVALID;
            }
        }


        if( ABBL_FLAGS(ibbl) & FB_ALIGNED_4 )
        {
            if( (offset & 0x3) != 0 )
            {
#if 1
                WARNING(("block %d not aligned: %d\n",ibbl,offset));
#else
                while( (offset & 0x3) != 0 )
                {
                    AsecWriteData32(isec,offset,0x2ffe0000);
                    offset++;
                    count++;
                    weight += ABBL_COUNT(ibbl);
                }
#endif
            }
        }

        if( ABBL_FLAGS(ibbl) & FB_ALIGNED_2 )
        {
            if( (offset & 1) != 0 )
            {
#if 1
                WARNING(("block %d not aligned: %d\n",ibbl,offset));
#else
                while( (offset & 1) != 0 )
                {
                    AsecWriteData32(isec,offset,0x2ffe0000);
                    offset++;
                    count++;
                    weight += ABBL_COUNT(ibbl);
                }
#endif
            }
        }

        
        ABBL_NEW_ADDR(ibbl) = offset;
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            /* remember the new index in the eval field */
            /* NEEDS WORK abuse of EVAL field */
        
            AINS_C_EVAL(iins) = offset;
            AsecWriteData32(isec,offset,Assemble(&AINS(iins)));
            offset++;
        }
        
        
        if( VALID(ABBL_EDG_FALSE(ibbl)) )
        {
            CASSERT( !VALID(ft_edge),("weird\n"));
            ft_edge = ABBL_EDG_FALSE(ibbl);
        }
    }
    
    
    /*
      compute new and old text section  size - this is also RELOCATED !!!
      we have already made the text section larger than it needs to be ! 
     */

    while( (offset & 0xf) != 0 )
    {
        AsecWriteData32(isec,offset,0x2ffe0000);
        offset++;
    }
    
    /*
    BIN_NEW_TEXT_SEC_SIZE = offset * 4; 
    */
    
    InfoStatus(fp,"\t[EmitGenerate] old text size: %ld  (%lx) new text size: %d (%x)  aligns %d wt %ld\n",
            AsecGetSize(isec), AsecGetSize(isec), offset * 4, offset * 4,count,weight);

    AsecSetSize(isec,offset * 4 );

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "TranslateAddress"

extern ADDRESS TranslateAddress(ADDRESS old)
{
    INDEX iins;
    INDEX ibbl;
    
    iins = Addr2Ins(old);
    CASSERT( AINS_IS_RELOC(iins),("target ins %d is not reloc (adr: %016lx)\n",iins,old));
    if( AINS_FLAGS(iins) & FI_LAST )
    {
        INDEX isec = SysGetTextSection();
        return Ins2Addr(AsecGetSize(isec)/4);
    }
    else
    {
        ibbl =  AINS_BBL(iins);
        CASSERT( ABBL_IS_RELOC(ibbl),("bbl %d is not reloc\n",ibbl));
        if( ABBL_IS_KILLED(ibbl) )
        {
            WARNING(("bbl %d is dead\n",ibbl));
            return 0;
        }
        else
        {
            return Ins2Addr(ABBL_NEW_ADDR(ibbl));
        }
        
    }
    

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "EmitRecomputeAddresses"

extern void EmitRecomputeAddresses(FILE *fp,INDEX xfun)
{
    int i;
    INDEX ibbl;
    INDEX isec;
    ADDRESS address;
    NUMBER count = 0;
    NUMBER bcount = 0;
    INDEX iins;
    NUMBER offset;
    INDEX  target;
    INDEX icount;
    
    FOREACH_FUN_BBL(xfun,ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_BRANCH )
        {
            target = AEDG_TAIL(ABBL_EDG_TRUE(ibbl)); 
        }
        else if( ABBL_TYPE(ibbl) == BT_UNCOND )
        {
            target = AEDG_TAIL(ABBL_EDG_UNCOND(ibbl));
        }
        else if( ABBL_TYPE(ibbl) == BT_EXTRA )
        {
            target = AEDG_TAIL(ABBL_EDG_EXTRA(ibbl)); 
        }
        else if( ABBL_TYPE(ibbl) == BT_CALL )
        {
            target = AEDG_TAIL(ABBL_EDG_CALL(ibbl));
        }
        else
        {
            continue;
        }

        /*
          the jump or branch instruction may not be the last
          instruction anymore because of the adding of noops
          */
        
        FOREACH_BBL_INS_R(ibbl,iins)
        {
            if( AINS_TYPE(iins) == IT_JMP ) break;
            if( AINS_TYPE(iins) == IT_IBR ) break;
            if( AINS_TYPE(iins) == IT_FBR ) break;
            if( AINS_TYPE(iins) == IT_BRA ) break;
        }

        CASSERT( VALID(iins),("not control instructions found\n"));

        if( AINS_TYPE(iins) == IT_JMP ) continue;
        
        bcount++;
                
        CWARNING( target < 2,("potential bogus target address for ins %d\n",iins));
        
        offset = ABBL_NEW_ADDR(target) - AINS_C_EVAL(iins) -1;
        if( ABBL_NEW_ADDR(target) < 2 )
        {
            ERROR(("bad new address for bbl: %d\n"));
            PrintBblSuper(stdout,target);
        }
        
        AINS_DATA(iins) = offset;
        if( offset == -1 )
        {
            PrintSep3(stdout);
            WARNING(("Endless loop encountered in bbl: %d\n",ibbl));
            PrintSep3(stdout);
            PrintEdg(stdout,ABBL_EDG_TRUE(ibbl));
            PrintIns(stdout,iins);            
            PrintBbl(stdout,ibbl);
            PrintBbl(stdout,target);

        }
    }
    
    /* create an ordered list of all relocatable addresses */
    
    SysRelocate(fp);
 
    /* recompute the gprel32 blocks at the beginning of the data segment */

    isec = SysGetSwitchSection();
#if 0
    printf("switch start is %lx\n", Global.SwitchStart);
    printf("section start is %lx\n", ASEC_START(isec));
#endif
    
    icount = (Global.SwitchStart - ASEC_START(isec))/4;
#if 0
    printf("icount is %d\n", icount);
#endif
    
    for(i=0;i<CONFIG_UTAB;i++)
    {
        if( AsecReadData32(isec,i+icount) == INVALID) continue;
        if( AsecReadData32(isec,i+icount) == 0) continue;

        address = SysGetGlobalPointer() + AsecReadData32(isec,i+icount);

        iins = Addr2Ins(address);
        
        /*
        printf("Data Before %08x ins: %d bbl: %d\n",ReadData32(code),iins,AINS_BBL(iins));
        */
        
        if( !(AINS_FLAGS(iins) & F_GPREL32) )
        {
            ERROR(("ins is not not FI_GPREL32\n"));
            PrintIns(stdout,iins);
        }
#if 0
        address = GetNewAddress(address,pair);
#else
        address = TranslateAddress(address);
#endif
        AsecWriteData32(isec,i+icount,(INT32)(address-SysGetGlobalPointer()));

        /*
        printf("Data After  %08x \n",ReadData32(code));
        */
    }
        
    /* new entry address */

    address = SysGetProgramEntry();

    CWARNING( address != TranslateAddress(address),("Entry address has changed\n"));
    
    /* new section size */
      
    isec = SysGetTextSection();

    SysUpdateTextSegmentSize();

    InfoStatus(fp,"\t[EmitAddress] locs: %d  brs: %d\n",count,bcount);
}

/*=======================================================
=======================================================*/








