/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.constants.c,v 1.5 1998/10/29 18:20:34 muth Exp muth $
  Comment:  This file contains the optimization code for constants
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

/*=======================================================
@OPTIMIZATION
Compute constants faster.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstantsOne"

extern BOOL OptConstantsOne(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Constone");
    BOOL change = 0;
    INDEX iins;
    NUMBER ccount=0;
    NUMBER ccounth=0;
    clock_t time = clock();
    
    FOREACH_INS(iins)
    {
        INT64 value;
        
        if( AINS_IS_KILLED(iins) ) continue;
        if( !(AINS_FLAGS(iins) & FI_EVAL_C) ) continue;
        if( AINS_REG_C(iins) == REG_ZERO ) continue;
        if( AINS_REG_C(iins) >= REG_FLOAT ) continue;
        
        if( AINS_TYPE(iins) == IT_LDA && AINS_REG_B(iins) == REG_ZERO ) continue;

        value = AINS_C_EVAL(iins);

        
        /* check whether it is in the text segement */

        if( AddressIsCode(value) )
        {
#if 0
            if( verbose )
            {
                fprintf(fp,"\t[*] ignoring code pointer: %016lx\n",value);
            }
#endif            
            continue;
        }
        
        
        if( -(1<<15) <=  value  && value < (1<<15) )
        {
            AinsMakeLda(iins,REG_ZERO,AINS_REG_C(iins),value);
            ccount++;    
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
                ccounth++;

            if( verbose )
            {
                fprintf(fp,"\t[*] ins %d create lda for value: %08lx\n",iins,value);
            }
            
            continue;
        }

        if( (value & 0xffff) == 0 && ((-(1L<<15))<<16) <= value && value < (1L<<31) )
        {
            AinsMakeLdah(iins,REG_ZERO,AINS_REG_C(iins),value>>16);
            ccount++;    
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
                ccounth++;

            if( verbose )
            {
                InfoExtra(fp,"\t[*] ins %d create ldah for value: %08lx\n",iins,value);
            }

            continue;
        }

        
        if( AINS_TYPE(iins) == IT_LDA )
        {
            if( AINS_REG_C(iins) == REG_GP ||  AINS_REG_B(iins) == REG_GP )
                continue;
        }


        value -= SysGetGlobalPointer();

        if( -(1<<15) <=  value  && value < (1<<15) )
        {
            AinsMakeLda(iins,REG_GP,AINS_REG_C(iins),value);
            ccount++;
            if( ABBL_IS_HOT(AINS_BBL(iins)) )
                ccounth++;

            if( verbose )
            {
                fprintf(fp,"\t[*] ins %d create gp-lda  for value: %016lx\n",iins,value);
            }
            
            continue;
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (ccount > 0);

    InfoStatus(fp,"\t[OptConstone%c] Optimized Constants: %d#%d {%d ms}\n",CHANGE(change),ccount,ccounth,time);

    return change;
}


/*=======================================================
@OPTIMIZATION
Sophisticated constant propagation.
We resuse a constant in some other register to (hopefully) reduce 
register usage.
This is only done on the basic block level.
We do not kill instructions but expect useless code to be eliminated
by other optimizations.
We also allow construction of values from other using
either explict lda's or doing it implicitly for
memory format instruction. Care must be taken to not construct
values that point into code sections since those values
may change!!!
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptConstantsTwo"

extern BOOL OptConstantsTwo(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Consttwo"); 
    BOOL   change = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    INDEX  ibbl;
    REG    reg;
    
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        INDEX  iins;
        
        REGLIST hasValue;
        INT64   CurrValue[64];
        NUMBER  LastDef[64];
        NUMBER  no;
        
        hasValue = REGS_NONE;
        for(reg=0;reg<64;reg++)
            LastDef[reg]=0;
        
        no = 0;

        
        FOREACH_BBL_INS(ibbl,iins)
        {
            REG ra = AINS_REG_A(iins);
            REG rb = AINS_REG_B(iins);
            REG rc = AINS_REG_C(iins);
            INT64 va = AINS_A_EVAL(iins);
            INT64 vb = AINS_B_EVAL(iins);
            INT64 vc = AINS_C_EVAL(iins);
            NUMBER oldest;
            REG    newreg;
            BOOL   isMemFormat;
            
            no++;

            /* REGISTER A */
            
            if( ra != REG_ZERO && ra != REG_LIT && REG_IS_INT(ra) && (AINS_FLAGS(iins) & FI_EVAL_A) )
            {
                oldest = no;
                newreg = ra;
                
                REGSET_FOREACH_REG(hasValue & REGS_INT,reg)
                {
                    if( CurrValue[reg] != va ) continue;
                    if( oldest > LastDef[reg] )
                    {
                        oldest = LastDef[reg];
                        newreg = reg;
                    }
                }
                    
                if( newreg != ra )
                {
                    if( verbose )
                    {
                        fprintf(fp, "\t[*] bbl: %d ins: %d rename reg a %d -> %d (%016lx)\n",ibbl,iins,ra,newreg,va);
                        PrintIns(stdout,iins);
                        /*
                        PrintBblSuper(stdout,AINS_BBL(iins));
                        */
                    }

                    AINS_REG_A(iins) = newreg;
                    if( ABBL_IS_HOT(ibbl) )
                        pcounth++;
                    else
                        pcount++;
                }
            }
#if 1
            /* REGISTER B (more oppportunities than with reg a because we can do an addition with
               a constant in the instruction) */
            
            if( rb != REG_ZERO && rb != REG_LIT && rb != REG_GP && REG_IS_INT(rb) && (AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                if( hasValue & REG2BIT(rb) )
                {
                    CWARNING(CurrValue[rb]!=vb,("rb %d new %lx old %lx\n",rb,CurrValue[rb],vb));
                }
                else
                {
                    hasValue |= REG2BIT(rb);
                    CurrValue[rb] = vb;
                }
                    
                oldest = no;
                newreg = rb;

                /*
                  we do not mess a lot with lda/ldah since if the base
                  register know the target register will be known as well
                   */
                
                if( (AINS_IS_LOAD(iins) || AINS_IS_STORE(iins)) && !AddressIsCode(vb) )
                    isMemFormat = 1;
                else
                    isMemFormat = 0;

                REGSET_FOREACH_REG(hasValue & REGS_INT,reg)
                {
                    if( isMemFormat )
                    {
                        if( vb + AINS_DATA(iins) - CurrValue[reg] >= (1<<15) ) continue;
                        if( vb + AINS_DATA(iins) - CurrValue[reg] <  -(1<<15) ) continue;
                    }
                    else
                    {
                        if( vb != CurrValue[reg] ) continue;
                    }
                    
                    if( oldest > LastDef[reg] )
                    {
                        oldest = LastDef[reg];
                        newreg = reg;
                    }
                }

                if( newreg != rb )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] bbl: %d ins: %d rename reg b %d (%016lx) -> %d (%016lx)\n",
                                ibbl,iins,rb,vb,newreg,CurrValue[newreg]);
                        PrintIns(stdout,iins);
                        /*
                        PrintBblSuper(stdout,AINS_BBL(iins));
                        */
                    }

                
                    if( isMemFormat )
                    {
                        AINS_DATA(iins) =  vb + AINS_DATA(iins) - CurrValue[newreg];
                        AINS_FLAGS(iins) &= ~FI_EVAL_B;
                        
                        AINS_B_EVAL(iins) = CurrValue[newreg];
                    }
                    
                    AINS_REG_B(iins) = newreg;
                    
                    if( ABBL_IS_HOT(ibbl) )
                        pcounth++;
                    else
                        pcount++;
                }
            }
#endif
             /* REGISTER C */

            if( rc != REG_ZERO && REG_IS_INT(rc) && (AINS_FLAGS(iins) & FI_EVAL_C) )
            {
#if 1
                if( !AddressIsCode(vc) )
                    isMemFormat = 1;
                else
                    isMemFormat = 0;

                oldest = no;
                newreg = INVALID;

                REGSET_FOREACH_REG(hasValue & REGS_INT,reg)
                {
                    if( isMemFormat )
                    {
                        INT64 difference =  vc - CurrValue[reg];
#if 0
                        if( (difference & 0xffff) == 0 && difference != 0 &&
                            ((-(1L<<15))<<16) <= difference && difference < (1L<<31) )
                        {
                            HURRAH(("Potential %lx %lx - %ld\n",vc,CurrValue[reg],ABBL_COUNT(ibbl)));
                        }
#endif                   
                            
                       
                        if( difference >= (1<<15) ) continue;
                        if( difference <  -(1<<15) ) continue;
                    }
                    else
                    {
                        if( vc != CurrValue[reg] ) continue;
                    }
                    
                    if( oldest > LastDef[reg] )
                    {
                        oldest = LastDef[reg];
                        newreg = reg;
                    }
                }

                if( VALID(newreg) && (AINS_TYPE(iins) != IT_LDA || (rb!=newreg && rb!=REG_GP && rb!=REG_ZERO) ))
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] bbl: %d  ins: %d rename reg c %d (%016lx) -> %d (%016lx)\n",
                                ibbl,iins,rc,vc,newreg,CurrValue[newreg]);
                        PrintIns(stdout,iins);
                        /*
                        PrintBblSuper(stdout,AINS_BBL(iins));
                        */
                    }
                    
                    if( isMemFormat )
                    {
                        AinsMakeLda(iins,newreg,rc, vc - CurrValue[newreg]);
                    }
                    else
                    {
                        AinsMakeLda(iins,newreg,rc, 0);
                    }

                    if( ABBL_IS_HOT(ibbl) )
                        pcounth++;
                    else
                        pcount++;
                }
#endif
                
                hasValue |= REG2BIT(rc);
                CurrValue[rc] = vc;
                LastDef[rc] = no;

            }
            else
            {
                hasValue &= ~REG2BIT(rc);
            }
        }
    }

               
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = pcount>0;

    InfoStatus(fp,"\t[OptConsttwo%c] Modified ins: %d#%d {%d ms}\n",CHANGE(change),pcount,pcounth,time);

    return change;
}

/*=======================================================
=======================================================*/
