/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.inline.c,v 1.38 1998/12/02 17:52:56 muth Exp muth $
  Comment:  This file contains the code for the optimizations:
                o inlining
                o pathinlining
=======================================================*/
#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

#define L1CACHE_SZ    2048    /* level-1 cache size: 8 Kbytes == 2 Kinstrs */

#define ABBL_CALL_IS_HOT ABBL_IS_HOT

/*=======================================================
=======================================================*/
static INDEX *SortedFun = 0;

#define SORTED_FUN(i)   SortedFun[i]

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "InitInline"

static void InitInline()
{
    /* the flag fields are the set to zero by calloc */
    CASSERT( SortedFun == 0,("weird\n"));
    SortedFun = (INDEX *)calloc(CONFIG_UFUN,sizeof(INDEX));
    CASSERT( SortedFun != 0,("Cannot Allocate Memory For SortedEdges\n"));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ExitInline"

static void ExitInline()
{
    CASSERT( SortedFun != 0,("weird\n"));
    free(SortedFun);
    SortedFun = 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "cmp"

static int cmp(const void *x1,const void *x2)
{
    INDEX e1 = *(INDEX *)x1;
    INDEX e2 = *(INDEX *)x2;
    
    if( e1 == INVALID && e2 == INVALID ) return 0;

    if( e1 == INVALID ) return 1;

    if( e2 == INVALID ) return -1;

    return AFUN_NUM_INSS(e1) - AFUN_NUM_INSS(e2);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunFlowFraction"     

static double AfunFlowFraction(INDEX callee,INDEX callsite)
{
    INDEX init = AFUN_BBL_FIRST(callee);
    double flow_frac;
    
    if( ABBL_COUNT(init) == 0 )
    {
        /*
        CWARNING(AFUN_IS_CRITICAL(callee),
                 ("Init bbl of crit fun has 0 count %d [%s]\n",callee,AFUN_NAME(callee)));*/
        flow_frac = 0.0;
    }
    else
    {
        flow_frac = (double) ABBL_COUNT(callsite) / (double)ABBL_COUNT(init);
        if( flow_frac < 0.0 ) flow_frac = 0.0;
        if( flow_frac > 1.0 ) flow_frac = 1.0;;
    }

    return flow_frac;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblFindCreateDummyEntryFun"

static  INDEX AbblFindCreateDummyEntryFun(INDEX xbbl)
{
    INDEX xfun;
    INDEX ifun;
    INDEX ibbl;
    INDEX iedg;
    INDEX iins;
    INDEX ret_save;
    REGLIST avail;
    REG call_reg;        
    REG ret_reg;


    
    CASSERT( !ABBL_IS_FIRST(xbbl),("Trying to create dummy entry for first bbl %d\n",xbbl));

    /* if entry fun already exists find and return it */

    xfun = ABBL_FUN(xbbl);
#if 0
    WARNING(("Looking for dummy entry %d %d[%s]\n",xbbl,xfun,AFUN_NAME(xfun)));
#endif

    FOREACH_BBL_PRED(xbbl,iedg)
    {
        ibbl = AEDG_HEAD(iedg);
        ifun = ABBL_FUN(ibbl);
        if( ifun == xfun ) continue;
        if( AFUN_MASTER(ifun) == xfun )
        {
#if 0
            WARNING(("found ibbl %d %d\n",ibbl,ifun));
#endif
            return ifun;
        }
        
    }

    /* if it does not exist create a new one */

    ifun = AfunNewFancy();
    ibbl = AbblNew();
    
    AFUN_NAME(ifun) = "Dummy";
    AFUN_MASTER(ifun) = xfun;
    AFUN_FLAGS(ifun) &= ~(FF_STK_INVALID|FF_STK_ARG|FF_STK_ALIAS);
    AFUN_FLAGS(ifun) |= AFUN_FLAGS(xfun) & (FF_STK_INVALID|FF_STK_ARG|FF_STK_ALIAS);
    AFUN_STK_MAX(ifun) = AFUN_STK_MAX(xfun);
    AFUN_STK_MIN(ifun) = AFUN_STK_MIN(xfun);
    CERROR( AFUN_HAS_STK_INVALID(ifun),("escape to function with bad stack\n"));
    
    
    ABBL_FUN(ibbl) = ifun;
    ABBL_TYPE(ibbl) = BT_NORM;
    AbblAppendFancy(ibbl);
    
    ret_reg = AfunReturnRegister(xfun);
    ret_save = AfunRegSaveIns(xfun,ret_reg);
#if 0
    WARNING(("creating dummy entry for %d   %d\n",xbbl,ibbl));
#endif
    if( VALID(ret_save) )
    {
        avail = ~AbblRegsLiveIn(xbbl) & 0x7fffffffL;
        
        if( avail == REGS_NONE )
        {
            ERROR(("could not find a call register\n"));
            PrintBblSuper(stdout,xbbl);
        }
        

        /* try to make the ret reg the call reg */
        if( avail & REG2BIT(ret_reg) )
        {
            call_reg = ret_reg;
        }
        else
        {
            call_reg = RegsGetFirst(avail);
#if 0
            WARNING(("using call reg: %d (ret reg %d) for bbl %d\n",call_reg,ret_reg,xbbl));
#endif
        }
#if 0
        WARNING(("creating save instruction call reg: %d return reg:  %d\n",call_reg,ret_reg));
#endif
        
        iins = AinsDup(ret_save);
        AINS_REG_A(iins) = call_reg;
        AINS_BBL(iins) = ibbl;
        AinsAppend(iins);
    }
    
    AedgLinkFancy( AedgNewFancy(ibbl,xbbl,ET_NORM) );
    return ifun;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunMustBeDuplicated"     

static BOOL AfunMustBeDuplicated(INDEX ifun)
{
    /*
      This implicitly catches reloc functions because
      usually they have more than one predecessor
      This need not be checked anymore since we
      now have additional edges from AFUN_JSR to each init bbl.

      If the function is jumped into this does not hurt the unlining
      as long as the function is duplicated 
     */
    

    
    if( AfunNCallSites(ifun) > 1 ) return 1;

    /*
      if a function is jumped into from somewhere else
      it is pretty much as good as a call from somewhere else 
      */
    
    if( AfunHas2Esc(ifun) ) return 1;

    return 0;
}

/*=======================================================
@FUN
Duplicate The Critical Path of a Function
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunDupCriticalPath"     

#define ABBL_MAGIC(i) ABBL_U1(i).doubleword.d1

extern INDEX AfunDupCriticalPath(INDEX ifun,double flow_frac)
{
    INDEX xfun;
    INDEX ibbl;
    INDEX jbbl;
    INDEX iedg;
    INDEX jedg;
    INDEX itab;
    INDEX jtab;
    BOOL  critpath_is_recursive = 0;
    REG  ret_reg = AfunReturnRegister(ifun);

#if 0
    REGLIST do_not_save;
    INDEX ret_save = AfunRegSaveIns(ifun,ret_reg);
    REG   reg;
#endif
    
#ifdef DBG_CP    
    printf("@@DUPLICATING:\n");
    PrintFunSuper(stdout,ifun);
    /*printf("@@RETURN reg: %d  off: %d\n",ret_reg,ret_off); */
#endif

    /* create a new function */
    
    xfun = AfunNewFancy();
    AFUN_NAME(xfun) = AFUN_NAME(ifun);
        
    /*
      duplicate critical bbls remember new bbl in magic
      field of corresponding old bbl
       */
    
       
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_IS_LAST(ibbl) )
        {
            ABBL_MAGIC(ibbl) = AFUN_BBL_LAST(xfun);
            continue;
        }

        if( !ABBL_IS_CRITICAL(ibbl) && !ABBL_IS_CRIT_PATH(ibbl) )
        {
            ABBL_MAGIC(ibbl) = INVALID;            
            continue;
        }
        
        
        jbbl = AbblDup(ibbl);
        ABBL_FUN(jbbl) = xfun;
        ABBL_MAGIC(ibbl) = jbbl;
        AbblAppendFancy(jbbl);

        if( ABBL_FLAGS(ibbl) & F_GPREL32 )
        {
            ABBL_FLAGS(jbbl) |= F_GPREL32;
            AINS_FLAGS(ABBL_INS_FIRST(jbbl)) |= F_GPREL32;
        }
        
        
        ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl) * flow_frac;
        ABBL_COUNT(ibbl) -= ABBL_COUNT(jbbl);
	
	if (ABBL_TYPE(ibbl) == BT_CALL && !critpath_is_recursive) {
	  FOREACH_BBL_SUCC(ibbl, iedg) {
	    if (AEDG_TYPE(iedg) == ET_FUN_CALL && AEDG_IS_RECURSIVE(iedg)) {
	      critpath_is_recursive = 1;
	    }
	  }
	}
    }

#if 0
    /*
      compute saved changed stuff, order is important 
     */

    AFUN_REGS_SAVED(xfun) = AFUN_REGS_SAVED(ifun) & ~REG2BIT(ret_reg);
    AfunComputeChangedRegisters(xfun);

    do_not_save = AFUN_REGS_SAVED(xfun) & ~AFUN_REGS_CHANGED(xfun);
    
    if( do_not_save )
    {
        printf("POTENTIAL SAVINGS: s %016lx c %016lx - %016lx\n",
               AFUN_REGS_SAVED(xfun),AFUN_REGS_CHANGED(xfun),do_not_save);

        for(reg=0;reg<64;reg++)
        {
            if( REG2BIT(reg) & do_not_save )
            {
                AfunKillLoadStoreSaveRegister(xfun,reg);
            }
        }
    }
#endif

    /*
      duplicate edges: we go backwards because we append
      some new bbls to function ifun
     */
        
    FOREACH_FUN_BBL_R(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) && !ABBL_IS_CRIT_PATH(ibbl) ) continue;
#ifdef DBG_CP    
        printf("@@DOING NODE: %d -> %ld\n",ibbl,ABBL_MAGIC(ibbl));
#endif        
        /* invariant: link edge before call edge */
        if( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL )
        {
#ifdef DBG_CP    
            printf("@@DOING CALL:\n");
#endif
            CASSERT( AbblNsuccs(ibbl) == 2,("bt_call has != 2 succs\n"));
            iedg = ABBL_EDG_LINK(ibbl);
            CASSERT( VALID(iedg),("invalid link_edge in bt_call\n"));
            CASSERT( ABBL_IS_CRITICAL(AEDG_TAIL(iedg)) || ABBL_IS_CRIT_PATH(AEDG_TAIL(iedg)),
                     ("et_fun_link is crit noncrit edge\n"));

            jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),
                                ABBL_MAGIC(AEDG_TAIL(iedg)),AEDG_TYPE(iedg));
            AedgLinkFancy(jedg);
	    AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);

            iedg = ABBL_EDG_CALL(ibbl);
            jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),
				AEDG_TAIL(iedg),AEDG_TYPE(iedg));
            AedgLinkFancy(jedg);
	    AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
        }
        else
        {
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                INDEX hd = AEDG_HEAD(iedg);
                INDEX tl = AEDG_TAIL(iedg);
                
                /* we do not want to create et_fun_return edges leading out of the last bbl */
                /* we do not want to create et_compensate edges leading out of the last bbl */
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
                if( AEDG_TYPE(iedg) == ET_COMPENSATE ) continue;
                if( AEDG_IS_INTER(iedg) )
                {
                    ERROR(("unexpected inter edge\n"));
                    PrintEdgSuper(stdout,iedg);
                }

#ifdef DBG_CP    
                printf("@@DOING EDGE (%c) %d -> %ld\n",AEDG_TYPE(iedg),tl,ABBL_MAGIC(tl));
#endif
                /*
                  PrintEdg(stdout,iedg);
                  */
                    
                if( !VALID(ABBL_MAGIC(tl)) )
                {
                    INDEX ifun_entry;
                    INDEX callsite;
                    INDEX returnsite;
                    INDEX dummy;
                    INDEX iins;

#ifdef DBG_CP    
                    printf("@@DOING CRTIT->NONCRIT:\n");
#endif
                    CERROR(ABBL_IS_CRIT_PATH(tl),("unexpected crit noncrit edge\n"));
                    CERROR(ABBL_IS_CRITICAL(tl),("unexpected crit noncrit edge\n"));

                    /* FIND/CREATE ENTRY FUN */

                    ifun_entry =  AbblFindCreateDummyEntryFun(tl);
                    dummy = AFUN_BBL_FIRST(ifun_entry);
                    
#if 0
printf("@@@---------- DUMMY ENTRY FUN:----------\n");
PrintFunSuper(stdout,ifun_entry);
printf("----------------------------------------\n");
#endif

                    /* CREATE CALL SITE */
                    
                    callsite = AbblNew();
                    ABBL_TYPE(callsite) = BT_CALL;
                    ABBL_FUN(callsite) = xfun;
                    ABBL_COUNT(callsite) = 0;
                    ABBL_FLAGS(callsite) |= FB_CRIT_ESC;
                    if( ABBL_FLAGS(tl) & F_GPREL32 )
                    {
                        ABBL_FLAGS(callsite) |= F_GPREL32;
                        AINS_FLAGS(ABBL_INS_FIRST(callsite)) |= F_GPREL32;
                    }
                    AbblAppendFancy(callsite);
                    
                    iins = AinsNew();
                    if( AbblNInss(dummy) > 0 )
                    {
                        AinsMakeBsr(iins,AINS_REG_A(ABBL_INS_FIRST(dummy)));
                        /* register might be patched later */
                    }
                    else
                    {
                        AinsMakeBsr(iins,AfunReturnRegister(ifun));  /* register might be patched later */
                    }
                    
                    AINS_BBL(iins) = callsite;
                    AinsAppend(iins);

                    /* CREATE RETURN SITE */
                    
                    returnsite = AbblNew();
                    ABBL_TYPE(returnsite) = BT_RET;
                    ABBL_FUN(returnsite) = xfun;
                    ABBL_COUNT(returnsite) = 0;
                    AbblAppendFancy(returnsite);
                    
                    iins = AinsNew();                /* will  be deleted very soon */
                    AinsMakeRet(iins,ret_reg);
                    AINS_BBL(iins) = returnsite;
                    AinsAppend(iins);

#if 0
                    /* CREATE DUMMY NODE */

                    dummy = AbblNew();
                    ABBL_TYPE(dummy) = BT_NORM;
                    ABBL_FUN(dummy) = ifun;
                    ABBL_COUNT(dummy) = 0;
                    AbblAppendFancy(dummy);
                    
                    if( VALID(ret_save) )
                    {
                        REGLIST avail = ~AbblRegsLiveIn(tl) & 0x7fffffffL;
                        REG call_reg;
                        
                        if( avail == REGS_NO )
                        {
                            ERROR(("could not find a call register\n"));
                            PrintBblSuper(stdout,tl);
                        }

                        if( avail & REG2BIT(ret_reg) )
                            call_reg = ret_reg;
                        else
                        {
                            call_reg = RegsGetFirst(avail);
#if 0
                            WARNING(("using call reg: %d (ret reg %d)\n",call_reg,ret_reg));
#endif
                        }
                        
                        /* NEEDS WORK: very bad hack: patch call register into bsr */

                        AINS_REG_C(ABBL_INS_FIRST(callsite)) = call_reg;

                        iins = AinsDup(ret_save);
                        AINS_REG_A(iins) = call_reg;
                        AINS_BBL(iins) = dummy;
                        AinsAppend(iins);
                    }

                    for(reg=0;reg<64;reg++)
                    {
                        if( !(REG2BIT(reg) & do_not_save) ) continue;

                        iins = AinsDup( AfunRegSaveIns(ifun,reg));
                        AINS_BBL(iins) = dummy;
                        AinsAppend(iins);
                    }
                    AedgLinkFancy( AedgNewFancy(dummy,tl,ET_NORM) );
#endif                    

                    AedgLinkFancy( AedgNewFancy(returnsite,returnsite,ET_EXIT_LINK) );
                    AedgLinkFancy( AedgNewFancy(callsite,returnsite,ET_FUN_LINK) );
                    AedgLinkFancy( AedgNewFancy(callsite,dummy,ET_FUN_CALL) );
                    ABBL_MAGIC(tl) = callsite;
                }

                AedgLinkFancy( AedgNewFancy(ABBL_MAGIC(hd),ABBL_MAGIC(tl),AEDG_TYPE(iedg)));                
            }
        }
    }

    FOREACH_FUN_BBL(xfun,ibbl)
    {
        ABBL_FLAGS(ibbl) |= FB_CRIT_PATH;
    }
    
    /* duplicate jump tables and patch the bbl*/

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !ABBL_IS_CRITICAL(ibbl) && !ABBL_IS_CRIT_PATH(ibbl) ) continue;
        if( ABBL_TYPE(ibbl) != BT_SWITCH ) continue;
        jbbl = ABBL_MAGIC(ibbl);
        CASSERT( VALID(ABBL_JUMP_TAB(ibbl)),("weird\n") );
        CASSERT( !VALID(ABBL_JUMP_TAB(jbbl)),("weird\n") );
        
        FOREACH_BBL_TAB(ibbl,itab)
        {
            jtab = AtabNew();
            ATAB_TARGET(jtab) = ABBL_MAGIC(ATAB_TARGET(itab));
            if( !VALID(ABBL_JUMP_TAB(jbbl)) )
                ABBL_JUMP_TAB(jbbl) = jtab;            
        }

        /* add terminator */
        jtab = AtabNew();
        ATAB_TARGET(jtab) = INVALID;

        AbblInstallNewJumpTable(jbbl);
    }
    
    if (critpath_is_recursive) {
      AFUN_IS_RECURSIVE(xfun) = 1;
    }

#ifdef DBG_CP
    printf("@@SAVINGS: %d - %d (flow %f)\n",AfunNInss(xfun),AfunNInss(ifun),flow_frac);
    printf("----------New (duplicated) function:----------\n");
    PrintFunSuper(stdout,xfun);
    printf("==============================================\n");
#endif

    return xfun;
}

/*=======================================================
@FUN
Duplicate (Clone) a Function
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunDup"

extern INDEX AfunDup(INDEX ifun, double flow_frac)
{
    /* magic field used to point from an original bbl to its copy */

    INDEX ibbl;
    INDEX jbbl;
    INDEX jfun;
    INDEX iedg;
    INDEX jedg;
    INDEX itab;
    INDEX jtab;
    
    CASSERT( flow_frac >= 0.0 && flow_frac <= 1.0,("bad flow fraction\n"));
    
    CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun), ("Bad argument fun %d\n",ifun));
    CASSERT( !AfunHasEsc2(ifun),("dup fun with esc2  %d [%s]\n",ifun,AFUN_NAME(ifun)));

    jfun = AfunNewFancy();
    AFUN_NAME(jfun) = AFUN_NAME(ifun) ;
    AFUN_FLAGS(jfun) = AFUN_FLAGS(ifun);
    AFUN_IS_RECURSIVE(jfun) = AFUN_IS_RECURSIVE(ifun);
    
    /* duplicate bbls and initialize their count fields */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_IS_LAST(ibbl) )
        {
            ABBL_MAGIC(ibbl) = AFUN_BBL_LAST(jfun);
            continue;
        }
        
        jbbl = AbblDup(ibbl);
        ABBL_FUN(jbbl) = jfun;
        ABBL_MAGIC(ibbl) = jbbl;
        AbblAppendFancy(jbbl);

        if( ABBL_FLAGS(ibbl) & F_GPREL32 )
        {
            ABBL_FLAGS(jbbl) |= F_GPREL32;
            AINS_FLAGS(ABBL_INS_FIRST(jbbl)) |= F_GPREL32;
        }
        
        ABBL_COUNT(jbbl) = (double) ABBL_COUNT(ibbl) * flow_frac;
        ABBL_COUNT(ibbl) -= ABBL_COUNT(jbbl);
    }

    /* duplicate edges */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        /* invariant: link edge before call edge */
        if( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL )
        {
            if( AbblNsuccs(ibbl) != 2 )
            {
                ERROR(("number of succs for call != 2\n"));
                PrintBblSuper(stdout,ibbl);
                ERROR(("number of succs for call != 2\n"));
            }
        
            iedg = ABBL_EDG_LINK(ibbl);
            jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),
                                ABBL_MAGIC(AEDG_TAIL(iedg)),AEDG_TYPE(iedg));
            AedgLinkFancy(jedg);
	    AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
            iedg = ABBL_EDG_CALL(ibbl);
            jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),AEDG_TAIL(iedg),AEDG_TYPE(iedg));
            AedgLinkFancy(jedg);
	    AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);
        }
        else
        {
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
                if( AEDG_TYPE(iedg) == ET_COMPENSATE ) continue;
                if( AEDG_IS_INTER(iedg) )
                {
                    ERROR(("unexpected inter edge: %d --> %d\n", AEDG_HEAD(iedg), AEDG_TAIL(iedg)));
                    PrintEdgSuper(stdout,iedg);
                }
                jedg = AedgNewFancy(ABBL_MAGIC(AEDG_HEAD(iedg)),
                                    ABBL_MAGIC(AEDG_TAIL(iedg)),AEDG_TYPE(iedg));
                AedgLinkFancy(jedg);
		AEDG_FLAGS(jedg) = AEDG_FLAGS(iedg);    
            }
        }
    }

    /* duplicate jump tables and patch the bbl*/

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_SWITCH ) continue;
        jbbl = ABBL_MAGIC(ibbl);
        CASSERT( VALID(ABBL_JUMP_TAB(ibbl)),("jump tab invalid\n") );
        CASSERT( !VALID(ABBL_JUMP_TAB(jbbl)),("jump tab valid\n") );
        
        FOREACH_BBL_TAB(ibbl,itab)
        {
            jtab = AtabNew();
            ATAB_TARGET(jtab) = ABBL_MAGIC(ATAB_TARGET(itab));
            if( !VALID(ABBL_JUMP_TAB(jbbl)) )
                ABBL_JUMP_TAB(jbbl) = jtab;            
        }

        /* add terminator */
        jtab = AtabNew();
        ATAB_TARGET(jtab) = INVALID;

        /*
        PrintBblSuper(stdout,jbbl);
        */
        AbblInstallNewJumpTable(jbbl);
        /*
        PrintBblSuper(stdout,jbbl);
        */
    }
    
            
    return jfun;
}

/*=======================================================
@FUNCTION
Inline a function "callee" at the basic block "callsite"
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunInline"

/* #define DO_SAVE_RESTORE */

extern void AfunInline(INDEX callee,INDEX callsite)
{
    INDEX caller = ABBL_FUN(callsite);
    INDEX init   = AFUN_BBL_FIRST(callee);
    INDEX exit   = AFUN_BBL_LAST(callee);
    INDEX returnsite = AEDG_TAIL(ABBL_EDG_LINK(callsite));
    INDEX ibbl;
    INDEX iedg;
    INDEX iins;
    REG ret_reg = AfunReturnRegister(callee);
#ifdef DO_SAVE_RESTORE
    REGLIST do_not_save;
#endif
    
    CWARNING( !VALID(ret_reg),("invalid return reg in %d [%s]\n",callee,AFUN_NAME(callee)));
    CASSERT( ABBL_TYPE(callsite) == BT_CALL,("bad callsite\n"));
    CASSERT( VALID(ABBL_EDG_LINK(callsite)),("no et_link in bt_call %d\n",callsite));
    CASSERT( VALID(returnsite),("invalid return site for bt_call %d\n",callsite));

#ifdef DO_SAVE_RESTORE
    do_not_save = AFUN_REGS_SAVED(callee) & ~AbblRegsLiveIn(returnsite);
#if 0
    printf("\t[*] %d [%s] s: %016lx",callee,AFUN_NAME(callee),AFUN_REGS_SAVED(callee));
#endif
    if( do_not_save )
    {
        REG reg;
        
        for(reg=0; reg<64; reg++)
        {
            if( REG2BIT(reg) & do_not_save)
            {
#if 0
                HURRAH(("Killing load store reg: %d\n",reg));
#endif
                AfunKillLoadStoreSaveRegister(callee,reg);
#if 0
                printf(" %d",reg);
#endif
            }
        }
    }
#if 0
    printf("\n");
#endif
#endif

    
#if 0
    printf("@@INLINEEE:\n");
    PrintFunSuper(stdout,callee);
#endif

    /* kill edges */
    
    AedgKillFancy(ABBL_EDG_CALL(callsite));  /* kill call and return edge */
    AedgKillFancy(ABBL_EDG_LINK(callsite));  /* kill link edge */


    /* inline bbls - so that we do not get interprocedural edges */

    FOREACH_FUN_BBL(callee,ibbl)
    {
        if( ABBL_IS_LAST(ibbl) ) continue;
        ABBL_FUN(ibbl) = caller;
    }

    /* insert init thru ABBL_PREV(exit) after callsite */
    
    ABBL_PREV(ABBL_NEXT(callsite)) = ABBL_PREV(exit);
    ABBL_NEXT(ABBL_PREV(exit)) = ABBL_NEXT(callsite);
    ABBL_NEXT(callsite) = init;
    ABBL_PREV(init) = callsite;

    /* make exit node the only node of callee */
    
    ABBL_PREV(exit) = ABBL_NEXT(exit) = INVALID;
    AFUN_BBL_FIRST(callee)  = AFUN_BBL_LAST(callee) = exit;


    /* do callsite bsr */

    iins = ABBL_INS_LAST(callsite);
    CASSERT( AINS_TYPE(iins) == IT_BRA,("not BRA\n"));
    CASSERT( AINS_CODE(iins) == I_bsr,("not br\n"));
    CWARNING( AINS_REG_C(iins) != ret_reg, ("bad ret reg %d(%d) in bsr %d [%s]\n",
                                            AINS_REG_C(iins),ret_reg,caller,AFUN_NAME(caller)));
    AinsKill(iins);
    ABBL_TYPE(callsite) = BT_NORM;
    AedgLinkFancy(AedgNewFancy(callsite,init,ET_NORM));

#ifdef DO_SAVE_RESTORE
    /* delete the instruction saving the return address. */
    FOREACH_BBL_INS_R(init, iins) {
      if (AINS_CODE(iins) == I_stq && AINS_REG_A(iins) == ret_reg) {
	AinsKill(iins);
	break;
      }
    }
#endif
    
    /* do exit ret's */

    CASSERT( ABBL_TYPE(exit) == BT_EXIT,("exit node is not bt_exit\n"));

    FOREACH_BBL_PRED(exit,iedg)
    {
        ibbl = AEDG_HEAD(iedg);
        iins = ABBL_INS_LAST(ibbl);
        
        CASSERT( ABBL_TYPE(ibbl) == BT_RET,("return node is not bt_ret\n"));
        CASSERT( AEDG_TYPE(iedg) == ET_EXIT_LINK,("non exit edge\n"));
        CASSERT( AINS_CODE(iins) == I_ret,("not ret\n"));
        CWARNING( AINS_REG_B(iins) != ret_reg,
                  ("bad ret reg %d(%d) in ret to %d [%s]\n",
                   AINS_REG_B(iins),ret_reg,caller,AFUN_NAME(caller)));

        AedgKillFancy(iedg);
        AinsKill(iins);

#if 0
	/* delete the instruction restoring the return address */
	FOREACH_BBL_INS_R(ibbl, iins)
        {  
            if( AINS_REG_C(iins) == ret_reg)
            {
                CASSERT( AINS_CODE(iins) == I_ldq,("weird return address restoring instruction\n"));
                AinsKill(iins);
                break;
            }
	}
#endif
        
        ABBL_TYPE(ibbl) = BT_NORM;
        AedgLinkFancy(AedgNewFancy(ibbl,returnsite,ET_NORM));
    }

    /* kill empty function */
    
    AbblKill(exit);
    AfunKill(callee);

#if 0
    printf("@@AFTER INLINING:\n");
    PrintFunSuper(stdout,caller);
#endif    
}
/*=======================================================
=======================================================*/

void PrintInlineInfo(FILE *fp,INDEX caller,INDEX callsite,INDEX callee,INDEX copy)
{
    fprintf(fp,"\t[*] %d [%s] -%ld-> %d [%s] (b %d i %d) was %d (b %d i %d s %d)\n",
            caller,AFUN_NAME(caller),ABBL_COUNT(callsite),
            copy,AFUN_NAME(copy),AfunNBbls(copy),AfunNInss(copy),callee,
            AfunNBbls(callee),AfunNInss(callee),-AFUN_STK_MIN(callee));
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunInlineSizeOK"

static BOOL AfunInlineSizeOK(INDEX callsite, INDEX callee, BOOL usepath)
{
  INDEX caller;
  INDEX topbbl, loophdr;
  INT32 currsz, calleesz;
  
  CASSERT((VALID(callsite) && ABBL_TYPE(callsite)==BT_CALL), 
	  ("bbl not a call site: %d\n", callsite));

  caller = ABBL_FUN(callsite);

  calleesz = usepath ? AfunNCriticalInss(callee) : AfunNInss(callee);

  if (VALID(ABBL_LOOP_HEADER(callsite))) {
    for (topbbl = callsite; VALID(ABBL_LOOP_HEADER(topbbl)) && ABBL_CALL_IS_HOT(topbbl);
	 topbbl = ABBL_LOOP_HEADER(topbbl)) {
      loophdr = ABBL_LOOP_HEADER(topbbl);
      currsz = usepath ? AbblLoopNCriticalInss(loophdr) : AbblLoopNInss(loophdr);
      if (currsz+calleesz > L1CACHE_SZ) {
	return FALSE;
      }
    }
  }
  else {
#if 0
    currsz = usepath ? AfunNCriticalInss(caller) : AfunFootprint(caller);
#endif
    currsz = usepath ? AfunNCriticalInss(caller) : AfunNInss(caller);
    if (currsz+calleesz > L1CACHE_SZ) {
      return FALSE;
    }
  }
  return TRUE;
}


/*=======================================================
@FUN
Returns TRUE if it is possible to reach callto from callfrom via a sequence of
function calls followed backwards, i.e., from callee to
caller.  This code is not very efficient, but can (and should) be made faster
if it's found to be a bottleneck.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblIsReachable"

extern BOOL AbblIsReachable(INDEX callfrom, INDEX callto)
{
  INDEX iedg, ifun, ibbl_entry;

  CASSERT(!(ABBL_IS_KILLED(callfrom) || ABBL_IS_KILLED(callto)), ("Weird\n"));

  if (ABBL_IS_MARKED(callfrom)) {
    return FALSE;
  }

  ifun = ABBL_FUN(callfrom);

  if (AFUN_IS_PSEUDO(ifun)) {
    return FALSE;     /* we won't ever inline this call */
  }

  if (ifun == ABBL_FUN(callto)) {
    return TRUE;
  }

  ABBL_MARK(callfrom);

  ibbl_entry = AFUN_BBL_FIRST(ifun);
  FOREACH_BBL_PRED(ibbl_entry, iedg) {
    if (AEDG_TYPE(iedg) != ET_FUN_CALL) {
      continue;
    }
    if (AbblIsReachable(AEDG_HEAD(iedg), callto)) {
      return TRUE;
    }
  }
  return FALSE;
}

/*=======================================================
@FUN
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AedgMarkRecursiveCalls"     

extern void AedgMarkRecursiveCalls()
{
  INDEX iedg, caller, callee;
  NUMBER nrec = 0, ncalls = 0;
  clock_t time = clock();

  FOREACH_LIVE_BBL(caller) {
    ABBL_UNMARK(caller);
  }

  FOREACH_LIVE_FUN(caller) {
      AFUN_IS_RECURSIVE(ABBL_FUN(caller)) = FALSE;
  }

  FOREACH_LIVE_EDG(iedg) {
    if (AEDG_TYPE(iedg) != ET_FUN_CALL) continue;

    AEDG_FLAGS(iedg) &= ~FE_RECURSIVE;

    caller = AEDG_HEAD(iedg);
    callee = AEDG_TAIL(iedg);

    if (AFUN_IS_PSEUDO(ABBL_FUN(caller)) || AFUN_IS_PSEUDO(ABBL_FUN(callee))) {
      continue;
    }

    ncalls++;

    if (!ABBL_CALL_IS_HOT(caller)) {
      continue;  /* call site is not hot and so won't be inlined */
    }

    if (AbblIsReachable(caller, callee)) {
      AEDG_FLAGS(iedg) |= FE_RECURSIVE;
      AFUN_IS_RECURSIVE(ABBL_FUN(caller)) = TRUE;
      AFUN_IS_RECURSIVE(ABBL_FUN(callee)) = TRUE;
      nrec++;
    }

    FOREACH_LIVE_BBL(caller) {
      ABBL_UNMARK(caller);
    }
  }
  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

  InfoStatus(stdout,"\t[MarkRecursiveCalls]: %d callsites, %d recursive {%d ms}\n",
	 ncalls, nrec, time);
}

/*=======================================================
@FUN
Counts the number of inlinable recursive functions.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunNInlinableRecursiveFns"     

static NUMBER AfunNInlinableRecursiveFns(BOOL usepath)
{
  INDEX ifun, ebbl, callsite;
  INDEX iedg;
  NUMBER n = 0;
  
  FOREACH_LIVE_FUN(ifun) {
    if (!AFUN_IS_RECURSIVE(ifun)) continue;
    if (AfunNotRecursivelyInlinable(ifun, usepath)) continue;

    ebbl = AFUN_BBL_FIRST(ifun);
    FOREACH_BBL_PRED(ebbl, iedg) {
      callsite = AEDG_HEAD(iedg);
      if (ABBL_TYPE(callsite) != BT_CALL) continue;
      if (!ABBL_IS_CRITICAL(callsite)) continue;
      n++;
    }
  }
  return n;
}

/*=======================================================
@FUN
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunWillNotBeInlined"     

extern BOOL AfunWillNotBeInlined(INDEX ifun,BOOL usepath)
{
    INDEX init;
    INDEX iedg;
    
    if( AFUN_FLAGS(ifun) & F_MARKED ) return 1;

    /*
      SKIP THOSE FUNCTIONS THAT WE CANNOT OR WANT NOT INLINE

      o NEEDS WORK: there are function that are not hot but
        have hot callsites
      o functions without hot bbls 
      o (really) bad stack manipulating functions (also takes care of next bullet) 
      o bad function like HELL,setjmp,longjmp, etc
      o fucntions with unresolved jumps
      o functions with escapes on critical path
      o functions with jumps on critical path
      */

    if( AfunNBblsHot(ifun) == 0 ) return 1;
    if( AFUN_HAS_STK_INVALID(ifun) ) return 1;
    if( AfunIsSetjmp(ifun) ) return 1;
    if( AfunIsLongjmp(ifun) ) return 1;
    if( AfunHasEsc2(ifun) ) return 1;

    init = AFUN_BBL_FIRST(ifun);
        
    FOREACH_BBL_PRED(init,iedg)
    {
        INDEX  callsite;
        INDEX  caller, callee;
        
        if( AEDG_TYPE(iedg) != ET_FUN_CALL ) continue;
        callsite = AEDG_HEAD(iedg);
        caller = ABBL_FUN(callsite);
        callee = ABBL_FUN(AEDG_TAIL(iedg));
        if( !ABBL_CALL_IS_HOT(callsite) ) continue;
        if( caller == ifun ) continue;
        /* we have a inlinable callsite !!! */
        if( VALID(INLINE_EXPANSION) )
        {
#if 0
            if( AfunNInss(caller)*100 > INLINE_EXPANSION * AFUN_NUM_INSS(caller) )
                continue;
#endif
	    if (AfunInlineSizeOK(callsite, callee, usepath)) continue;
        }
        return 0;
    }

    return 1;
}

/*=======================================================
@FUN
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblLoopNCriticalInss"     

extern INT32 AbblLoopNCriticalInss(INDEX loophdr)
{
  INDEX ibbl;
  INT32 sz = 0;

  CASSERT( (VALID(loophdr) && ABBL_IS_LOOP_HEADER(loophdr)), ("Weird\n"));
  ABBL_FOREACH_LOOP_MEMBER(loophdr, ibbl) {
    if (ABBL_IS_CRITICAL(ibbl)) {
      sz += AbblNInss(ibbl);
    }
  }
  return sz;
}

/*=======================================================
@FUN
Computes the total number of instructions in a loop.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AbblLoopNInss"     

extern INT32 AbblLoopNInss(INDEX loophdr)
{
  INDEX ibbl;
  INT32 sz;

  CASSERT( (VALID(loophdr) && ABBL_IS_LOOP_HEADER(loophdr)), ("Weird\n"));

  sz = AbblNInss(loophdr);
  ABBL_FOREACH_LOOP_MEMBER(loophdr, ibbl) {
    if (ABBL_IS_LOOP_HEADER(ibbl)) {
      sz += AbblLoopNInss(ibbl);
    }
    else {
      sz += AbblNInss(ibbl);
    }
  }
  return sz;
}

/*=======================================================
@FUN
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunNCriticalInss"     

extern INT32 AfunNCriticalInss(INDEX ifun)
{
  INDEX ibbl;
  INT32 sz = 0;

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_IS_CRITICAL(ibbl)) {
      sz += AbblNInss(ibbl);
    }
  }
  return sz;
}

/*=======================================================
@FUN
Computes the "cache footprint" of a function, i.e., the total no.
of instructions in the strongly connected component to which that
function belongs (for a nonrecursive function this is a singleton).
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunFootprint"     

static INT32 AfunAuxFootprint(INDEX ifun)
{
  INDEX jfun, ibbl, iedg;
  INT32 footprint;

  CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun), ("Bad argument fun %d\n",ifun));
  
  if (AFUN_IS_PSEUDO(ifun)) return 0;
  if (ABBL_IS_MARKED(AFUN_BBL_FIRST(ifun))) return 0;

  footprint = AfunNInss(ifun);
  if (!AFUN_IS_RECURSIVE(ifun)) {
    return footprint;
  }
  
  ABBL_MARK(AFUN_BBL_FIRST(ifun));

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) != BT_CALL) continue;
    
    FOREACH_BBL_SUCC(ibbl, iedg) {
      if (AEDG_TYPE(iedg) != ET_FUN_CALL) continue;
      if (!AEDG_IS_RECURSIVE(iedg)) continue;

      jfun = ABBL_FUN(AEDG_TAIL(iedg));
      footprint += AfunAuxFootprint(jfun);
    }
  }
  return footprint;
}

extern INT32 AfunFootprint(INDEX ifun)
{
  INDEX jfun;
  INT32 footprint;

  CASSERT( VALID(ifun) && !AFUN_IS_KILLED(ifun), ("Bad argument fun %d\n",ifun));
  
  if (AFUN_IS_RECURSIVE(ifun)) {
    FOREACH_LIVE_FUN(jfun) {
      if (AFUN_IS_PSEUDO(jfun)) continue;
      ABBL_UNMARK(AFUN_BBL_FIRST(jfun));
    }
  
    footprint = AfunAuxFootprint(ifun);
  }
  else {
    footprint = AfunNInss(ifun);
  }
  return footprint;
}    

/*=======================================================
@FUN
Returns TRUE if function callee should not be considered for recursive
inlining.  At this time, a recursive function callee is not considered
for inlining if any of the following hold:

(1) callee is a pseudo-function (shouldn't happen, but included for safety);

(2) callee contains a setjmp or longjmp;

(3) callee does not have a "good" callsite (hence, noweher for it to be inlined to);

(4) callee contains an indirect jump;

(5) callee contains a jump through a jump table (i.e., a switch);

(6) callee contains an "escaping" edge to another function;

If the boolean argument usepath is TRUE, only the critical path is considered.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunNotRecursivelyInlinable"

extern BOOL AfunNotRecursivelyInlinable(INDEX callee, BOOL usepath)
{
  INDEX ibbl, iedg;

  if( AFUN_IS_PSEUDO(callee) ) return TRUE;
  if( AfunIsSetjmp(callee) ) return TRUE;
  if( AfunIsLongjmp(callee) ) return TRUE;

  FOREACH_FUN_BBL(callee, ibbl) {
    if (usepath) {
      if (!ABBL_IS_CRITICAL(ibbl) && !ABBL_IS_CRIT_PATH(ibbl)) continue;
    }
    if (ABBL_TYPE(ibbl) == BT_JUMP || ABBL_TYPE(ibbl) == BT_SWITCH) {
      return TRUE;
    }
    if (!ABBL_IS_LAST(ibbl)) {
      FOREACH_BBL_SUCC(ibbl, iedg) {
	if(AEDG_IS_INTER(iedg) 
	   && AEDG_TYPE(iedg) != ET_FUN_RETURN
	   && AEDG_TYPE(iedg) != ET_FUN_CALL) {
	  return TRUE;
        }
      }
    }
  }
  return FALSE;
}

static struct dupmapstr {
    INDEX old, new;  /* keeps track of old and duplicated versions of */
  } *dupmap;         /* recursive functions, for redirecting calls */

#if 0
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunRedirectRecursiveEdges"

static void AfunRedirectRecursiveEdges(INDEX ifun, NUMBER nrecfns, BOOL usepath)
{
  INDEX ibbl, iedg, iedgnew;
  INDEX callee;
  int   j;

  FOREACH_FUN_BBL(ifun, ibbl) {
    if (ABBL_TYPE(ibbl) != BT_CALL) continue;
    FOREACH_BBL_SUCC(ibbl, iedg) {
      if ( !(AEDG_TYPE(iedg) == ET_FUN_CALL && AEDG_IS_RECURSIVE(iedg)) )
	continue;
      callee = ABBL_FUN(AEDG_TAIL(iedg));
      if (!AFUN_IS_RECURSIVE(callee)) continue;
      if (AfunNotRecursivelyInlinable(callee, usepath)) continue;

      /* edge should be redirected: find the "new version" of the callee */
      for (j = 0; j < nrecfns; j++) {
	if (callee == dupmap[j].old) {
	  break;
	}
      }
      CASSERT(j < nrecfns, ("Callee not found for redirection: %d [%s]\n", callee, AFUN_NAME(callee)));

      /* redirect the edge */
      AedgKillFancy(iedg);
      iedgnew = AedgNewFancy(ibbl,AFUN_BBL_FIRST(dupmap[j].new), ET_FUN_CALL);
      AedgLinkFancy(iedgnew);
      AEDG_FLAGS(iedgnew) |= FE_RECURSIVE;
    }
  }
}
#endif

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptInlineRecursive"

static BOOL OptInlineRecursive(FILE *fp, BOOL usepath)
{
  BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Inline");
  BOOL   verboserec = HasOption(PRINT_OPT_VERBOSE,"InlineRec");
  BOOL   inlined, has_crit_callsite;
  BOOL   change;
  NUMBER nrecfns, ninline_limit, nrounds = 0;
  NUMBER fcount=0, bcount=0, icount=0;
  INDEX  ifun, jfun, caller, callee;
  INDEX  ifun_entry, ibbl, jbbl;
  INDEX  iedg, jedg;
  int    i, j;
  double flowfrac;
  clock_t time = clock();

  AedgMarkRecursiveCalls();

  nrecfns = AfunNInlinableRecursiveFns(usepath);
  if (verboserec) {
    printf("\t\t[OptInlineRec: %d inlinable recursive fns]\n", nrecfns);
  }

  if (nrecfns == 0) {
    return 0;
  }
  dupmap = (struct dupmapstr *)calloc(nrecfns, sizeof(*dupmap));
  /*
   * first, duplicate each recursive function that can be duplicated, and 
   * record the information about the index of the duplicated version of 
   * each such function in the array pointed at by dupmap.  The functions
   * are processed in reverse order in order to avoid re-duplication of
   * the functions introduced by this duplication process (and added at
   * the end of the list of live functions).
   */
  i = 0;
  FOREACH_LIVE_FUN_R(ifun) {
    if (!AFUN_IS_RECURSIVE(ifun)) continue;
    if (AfunNotRecursivelyInlinable(ifun, usepath)) continue;

    ibbl = AFUN_BBL_FIRST(ifun);
    has_crit_callsite = FALSE;
    FOREACH_BBL_PRED(ibbl, iedg) {
      jbbl = AEDG_HEAD(iedg);
      if (ABBL_TYPE(jbbl) != BT_CALL) continue;
      if (!ABBL_IS_CRITICAL(jbbl)) continue;
      has_crit_callsite = TRUE;
      break;
    }
    if (!has_crit_callsite) continue;

    flowfrac = 0.0;
    ifun_entry = AFUN_BBL_FIRST(ifun);
    FOREACH_BBL_PRED(ifun_entry, iedg) {
      if (AEDG_IS_RECURSIVE(iedg)) {
	flowfrac += AfunFlowFraction(ifun, AEDG_HEAD(iedg));
      }
    }

    jfun = usepath ?
              AfunDupCriticalPath(ifun,flowfrac) :
              AfunDup(ifun,flowfrac);
    dupmap[i].old = ifun;
    dupmap[i].new = jfun;
    if (verboserec) {
      printf("OptInlineRec: duplicating fun %d, copy = %d [i = %d]\n",
	     ifun, jfun, i);
    }
    i++;
  }
  /*
   * inline (repeatedly, until no new inlining occurs: usually this
   * will occur because further inlining would cause the function body to
   * overflow the cache).
   */
  ninline_limit = 10000;   /* limit on no. of rounds of inlinings allowed */
  inlined = TRUE;
  for (nrounds = 0; inlined && nrounds < ninline_limit; nrounds++) {
    inlined = FALSE;
    for (i = 0; i < nrecfns; i++) {
      caller = dupmap[i].old;
      FOREACH_FUN_BBL(caller, ibbl) {
	ABBL_MARK(ibbl);
      }
      FOREACH_FUN_BBL(caller, ibbl) {
	if (!ABBL_IS_MARKED(ibbl)) continue;
	if (ABBL_TYPE(ibbl) != BT_CALL) continue;
	if (!ABBL_IS_CRITICAL(ibbl)) continue;
	FOREACH_BBL_SUCC(ibbl, iedg) {
	  if (AEDG_TYPE(iedg) != ET_FUN_CALL) continue;
	  if (!AEDG_IS_RECURSIVE(iedg)) continue;
	  callee = ABBL_FUN(AEDG_TAIL(iedg));
	  if (AfunNotRecursivelyInlinable(callee, usepath)) continue;
	  if (!AfunInlineSizeOK(ibbl, callee, usepath)) {
	    continue;
	  }
	  /* find saved copy of the callee */
	  for (j = 0; j < nrecfns; j++) {
	    if (callee == dupmap[j].old) {
	      break;
	    }
	  }
#if 1
	  if (j >= nrecfns) {
	  CASSERT(j < nrecfns,
		  ("Callee not found: %d [%s] (caller = %d [%s])\n", callee, AFUN_NAME(callee), caller, AFUN_NAME(caller)));
	  exit(0);
	  }
#endif
	  CASSERT(j < nrecfns,
		  ("Callee not found: %d [%s] (caller = %d [%s])\n", callee, AFUN_NAME(callee), caller, AFUN_NAME(caller)));

	  ifun = AfunDup(dupmap[j].new, AfunFlowFraction(callee,ibbl));
	  /* redirect the edge */
	  AedgKillFancy(iedg);
	  jedg = AedgNewFancy(ibbl,AFUN_BBL_FIRST(ifun), ET_FUN_CALL);
	  AedgLinkFancy(jedg);
	  AEDG_FLAGS(jedg) |= FE_RECURSIVE;
	  bcount+=AfunNBbls(ifun);
	  icount+=AfunNInss(ifun);
	  fcount+=1;
	  
	  if( verbose ) {
	    PrintInlineInfo(fp,caller,ibbl,callee,ifun);
	  }

	  AfunInline(ifun, ibbl);
	  inlined = TRUE;
	}
      }
    }
  }

  time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
  change = (fcount>0);

  InfoStatus(fp,"\t[OptInlineRec%s%c] Inlined Funs: %d  Bbls: %d  Inss: %d {%d ms}\n",
	 INLINE_METHOD,CHANGE(change),fcount,bcount,icount,time);
  return change;
}

/*=======================================================
@OPT
Inline Frequently Called Functions.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptInline"     

extern BOOL OptInline(FILE *fp,BOOL usepath )
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Inline");
    BOOL   change;
    NUMBER fcount=0;
    NUMBER bcount=0;
    NUMBER icount=0;
    INDEX  init;
    INDEX  callee;
    INDEX  ifun;
    INDEX  iedg;
    INDEX  i;
    clock_t time = clock();
    
    if( !CONFIG_USE_PROFILE )
      {
        printf("\t[OptInline] No Inlining because of lack of profiling information\n");
        return 0;
    }
    
    InitInline();
    
    if( verbose )
        printf("\t[*] Callsites Before Direct %d/%d  Indirect: %d/%d\n",
               AbinNHotCallsites(),AbinNCallsites(),AbinNHotIndirectCallsites(),AbinNIndirectCallsites());

#if 0
    if( VALID(INLINE_EXPANSION) )
    {
        WARNING(("INLINE EXPANSION MAX: %d%%\n",INLINE_EXPANSION));
    }
#endif

    FOREACH_LIVE_FUN(ifun)
    {
        AFUN_NUM_INSS(ifun) = AfunNInss(ifun);
        AFUN_FLAGS(ifun) &= ~F_MARKED;
    }

    for(ifun=0; ifun < CONFIG_UFUN; ifun++ )
        SORTED_FUN(ifun) = INVALID;
    
    FOREACH_LIVE_FUN(ifun)
    {
        SORTED_FUN(ifun) = ifun;
    }
    
    qsort(SortedFun,CONFIG_UFUN,sizeof(INDEX),cmp);
    
    /* create chains by decreasing edge count */

    for(i=0; i < CONFIG_UFUN; i++ )
    {
        NUMBER bottom_count = 0;
        NUMBER inline_count = 0;

        callee = SORTED_FUN(i);
        
        FOREACH_LIVE_FUN(callee)
        {
            if( AfunWillNotBeInlined(callee,usepath) )
            {
                if( !(AFUN_FLAGS(callee) & F_MARKED) )
                {
                    AFUN_FLAGS(callee) |= F_MARKED;
                    bottom_count++;
                }

                continue;
            }
            
            /* NOW INLINE */

            init = AFUN_BBL_FIRST(callee);
            
            
            FOREACH_BBL_PRED(init,iedg)
            {
                INDEX  callsite;
                INDEX  caller, callee;
		INT32  maxsz, currsz;

                if( AEDG_TYPE(iedg) != ET_FUN_CALL ) continue;
                callsite = AEDG_HEAD(iedg);
                caller = ABBL_FUN(callsite);
		callee = ABBL_FUN(AEDG_TAIL(iedg));
                if( !ABBL_CALL_IS_HOT(callsite) ) continue;
                if( !(AFUN_FLAGS(caller) & F_MARKED) ) continue;
                if( caller == callee ) continue;
#if 0
                if( VALID(INLINE_EXPANSION) )
                {
                    if( AfunNInss(caller)*100 > INLINE_EXPANSION * AFUN_NUM_INSS(caller) )
                        continue;
                }
#endif

		if (VALID(ABBL_LOOP_HEADER(callsite))) {
		  currsz = AbblLoopNCriticalInss(ABBL_LOOP_HEADER(callsite));
		}
		else {
		  currsz = AfunNCriticalInss(caller);
		}
		maxsz =  L1CACHE_SZ;
		if ( AFUN_NUM_INSS(callee)+currsz > maxsz ) {
		  continue;
		}

                if( AfunMustBeDuplicated(callee) )
                {
                    ifun = usepath ?
                           AfunDupCriticalPath(callee,AfunFlowFraction(callee,callsite)) :
                           AfunDup(callee,AfunFlowFraction(callee,callsite));
                    
                    bcount+=AfunNBbls(ifun);
                    icount+=AfunNInss(ifun);
                    fcount+=1;
                }
                else
                {
                    ifun = callee;
                    fcount+=1;
                }


                if( verbose )
                {
                    PrintInlineInfo(fp,caller,callsite,callee,ifun);
                }

                AfunInline(ifun,callsite);
                inline_count++;
            }
        }

#if 0
        printf("%d %d\n",inline_count,bottom_count);
#endif        
        if( inline_count == 0 && bottom_count == 0)
        {
            INDEX smallest=INVALID;
            
            FOREACH_LIVE_FUN(ifun)
            {
                if( AFUN_FLAGS(ifun) & F_MARKED ) continue;
                if( !VALID(smallest) ||
                    ABBL_COUNT(AFUN_BBL_FIRST(smallest)) > ABBL_COUNT(AFUN_BBL_FIRST(ifun)) )
                    smallest = ifun;
            }

            if( VALID(smallest) )
            {
                AFUN_FLAGS(smallest) |= F_MARKED;
#if 0
                WARNING(("forced function %d [%s] into bottom\n",
                         smallest,AFUN_NAME(smallest)));
#endif
            }
            else
            {
                break;
            }
        }
        
    }

    if( verbose )
        printf("\t[*] Callsites After Direct %d/%d  Indirect: %d/%d\n",
               AbinNHotCallsites(),AbinNCallsites(),AbinNHotIndirectCallsites(),AbinNIndirectCallsites());

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (fcount>0);

    InfoStatus(fp,"\t[OptInline%s%c] Inlined Funs: %d  Bbls: %d  Inss: %d {%d ms}\n",
           INLINE_METHOD,CHANGE(change),fcount,bcount,icount,time);

    change |= OptInlineRecursive(fp, usepath);

    ExitInline();
        
    return change;
}


/*=======================================================
@OPT
Inline Small Functions.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptInlineSmall"     

extern BOOL OptInlineSmall(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Inlinesmall");
    BOOL   change;
    NUMBER fcount=0;
    NUMBER bcount=0;
    NUMBER icount=0;
    INDEX  init;
    INDEX  callee;
    INDEX  callsite;
    INDEX  caller;
    INDEX  ifun;
    INDEX  iedg;
    clock_t time = clock();

    FOREACH_LIVE_FUN(ifun)
    {
        init = AFUN_BBL_FIRST(ifun);
        
        if( AFUN_IS_PSEUDO(ifun) ) continue;
        if( AfunIsSetjmp(ifun) ) continue;
        if( AfunIsLongjmp(ifun) ) continue;
        if( AfunHasEsc2(ifun) ) continue;
        if( !AfunIsLeaf(ifun) ) continue;
        if( AfunNInss(ifun) * AbblNpreds(init) > 64 && AfunNInss(ifun) > 4 ) continue;
        if( AFUN_STK_MIN(ifun) != 0 ) continue;
        if( AFUN_STK_MAX(ifun) != 0 ) continue;
        if( AfunNJumps(ifun) > 0 ) continue;
        if( AfunNSwitches(ifun) > 0 ) continue;
        
        FOREACH_BBL_PRED(init,iedg)
        {
            if( AEDG_TYPE(iedg) != ET_FUN_CALL ) continue;
            callsite = AEDG_HEAD(iedg);
            caller = ABBL_FUN(callsite);
            if( AFUN_IS_PSEUDO(caller) ) continue;

            callee = AfunDup(ifun,AfunFlowFraction(ifun,callsite));
            bcount+=AfunNBbls(callee);
            icount+=AfunNInss(callee);
            fcount++;
            if( verbose )
            {
                PrintInlineInfo(fp,caller,callsite,ifun,callee);
            }
            AfunInline(callee,callsite);
        }
    }
    
    

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (fcount>0);

    InfoStatus(fp,"\t[OptInlinesmall%c] Inlined Funs: %d  Bbls: %d  Inss: %d {%d ms}\n",
           CHANGE(change),fcount,bcount,icount,time);

    return change;
}

/*=======================================================
@OPT
Inline functions with single callsite.
This should always be a win with the exception
the inlining a function with bad stack behaviour will also
render the calling function to have bad stack behaviour
and vice versa. We ignore this problem.
But doing this optimization as late as possible will
ensure that other optimization that benefit from
well behaved stacks (load avaoidance) have already occured.
Note that sometimes inlining might result in the
stack behaviour being improved, eg. when pointers into the stack 
are passed as arguments.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptInlineOneCall"     

extern BOOL OptInlineOneCall(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Inlineonecall");
    BOOL   change;
    NUMBER fcount=0;
    INDEX  init;
    INDEX  callsite;
    INDEX  caller;
    INDEX  ifun;
    INDEX  iedg;
    
    clock_t time = clock();

    FOREACH_LIVE_FUN(ifun)
    {
        init = AFUN_BBL_FIRST(ifun);
        
        if( AFUN_IS_PSEUDO(ifun) ) continue;
        if( AfunIsSetjmp(ifun) ) continue;
        if( AfunIsLongjmp(ifun) ) continue;
        if( AfunHasEsc2(ifun) ) continue;
        if( AfunHas2Esc(ifun) ) continue;
        if( AfunNJumps(ifun) > 0 ) continue;
        if( AbblNpreds(init) != 1 ) continue;

        
        iedg = ABBL_PRED(init);
        if( AEDG_TYPE(iedg) != ET_FUN_CALL ) continue;
        
        callsite = AEDG_HEAD(iedg);
        caller = ABBL_FUN(callsite);
        if( AFUN_IS_PSEUDO(caller) ) continue;
#if 0
        printf("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$\n");
        
        PrintFun(stdout,caller);
        PrintFun(stdout,ifun);
#endif            
#if 1
        fcount++;
        if( verbose )
        {
            PrintInlineInfo(fp,caller,callsite,ifun,ifun);
        }
        AfunInline(ifun,callsite);
#endif
    }
    
    

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (fcount>0);

    InfoStatus(fp,"\t[OptInlineonecall%c] Inlined Funs: %d {%d ms}\n", CHANGE(change),fcount,time);

    return change;
}

/*=======================================================
=======================================================*/





