/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.normalize.c,v 1.1 1998/09/08 22:45:41 muth Exp muth $
  Comment:  This file contains the optimization code
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"


/*=======================================================
@OPTIMIZATION 
Normalize Instructions.
Often there are several instruction which achieve the same
effect eg, lda r0,5(r1) or add r1,5,r0.
In order to reduce the number of cases to be considered by
other optimizations we normalize instructions.
Those normlization should be (but aren't) undone before
scheduling picking the instruction which has the nicer
scheduling behaviour, eg. use bis r5,r5,r6 instead
of lda r6,0(r5)
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptNormalize"

extern BOOL OptNormalize(FILE *fp)
{
    /*
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Normalizeone");
    */
    BOOL   change;
    INDEX iins;
    NUMBER ncount=0;
    NUMBER ncounth=0;
    clock_t time = clock();
    
    /* convert addq,subq,s4addq,s8addq to lda, this simplifies stack management later*/
    
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_CODE(iins) == I_addq )
        {
            if( AINS_REG_B(iins)==REG_LIT )
            {
                AinsMakeLda(iins,AINS_REG_A(iins),AINS_REG_C(iins),AINS_DATA(iins));
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_B) &&
                     AINS_B_EVAL(iins) < (1<<15) &&
                     AINS_B_EVAL(iins) >= -(1<<15) )
                     
            {
#if 0                
                HURRAH(("simplified addq %d\n",AINS_B_EVAL(iins)));
                PrintIns(stdout,iins);
#endif                
                AinsMakeLda(iins,AINS_REG_A(iins),AINS_REG_C(iins),AINS_B_EVAL(iins));
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_A) &&
                     AINS_A_EVAL(iins) < (1<<15) &&
                     AINS_A_EVAL(iins) >= -(1<<15) )
            {
#if 0
                HURRAH(("simplified addq %d\n",AINS_A_EVAL(iins)));
                PrintIns(stdout,iins);
#endif
                AinsMakeLda(iins,AINS_REG_B(iins),AINS_REG_C(iins),AINS_A_EVAL(iins));
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
        }

        if( AINS_CODE(iins) == I_subq )
        {
            if( AINS_REG_B(iins)==REG_LIT )
            {
                AinsMakeLda(iins,AINS_REG_A(iins),AINS_REG_C(iins),-AINS_DATA(iins));
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
            else if( (AINS_FLAGS(iins) & FI_EVAL_B) &&
                     AINS_B_EVAL(iins) <= (1<<15) &&
                     AINS_B_EVAL(iins) > -(1<<15) )
            {
#if 0
                HURRAH(("simplified subq %d\n",AINS_B_EVAL(iins)));
                PrintIns(stdout,iins);
#endif           
                AinsMakeLda(iins,AINS_REG_A(iins),AINS_REG_C(iins),-AINS_B_EVAL(iins));
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
        }
        
        if( AINS_CODE(iins) == I_s4addq )
        {
            if( (AINS_FLAGS(iins) & FI_EVAL_A) &&
                AINS_A_EVAL(iins) < (1<<13) &&
                AINS_A_EVAL(iins) >= -(1<<13) )
            {
#if 0
                HURRAH(("simplified s4addq %d\n",AINS_A_EVAL(iins)));
                PrintIns(stdout,iins);
#endif
                AinsMakeLda(iins,AINS_REG_B(iins),AINS_REG_C(iins),AINS_A_EVAL(iins)*4);
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
            
        }

        if( AINS_CODE(iins) == I_s8addq )
        {
            if( (AINS_FLAGS(iins) & FI_EVAL_A) &&
                AINS_A_EVAL(iins) < (1<<12) &&
                AINS_A_EVAL(iins) >= -(1<<12) )
            {
#if 0
                HURRAH(("simplified s8addq %d\n",AINS_A_EVAL(iins)));
                PrintIns(stdout,iins);
#endif                
                AinsMakeLda(iins,AINS_REG_B(iins),AINS_REG_C(iins),AINS_A_EVAL(iins)*8);
                ncount++;
                if( ABBL_IS_HOT(AINS_BBL(iins)) )
                    ncounth++;
            }
        }

    }

    /* normalize moves */
    
    FOREACH_LIVE_INS(iins)
    {
        if( !AinsIsMove(iins) ) continue;
        if( AINS_CODE(iins) == I_lda ) continue;
        if( AINS_CODE(iins) == I_cpys ) continue;

        ncount++;
        if( ABBL_IS_HOT(AINS_BBL(iins)) )
            ncounth++;
                
        /* moves never have a lit register !!!! */
        
        if( AINS_REG_C(iins) >= 32 )
        {
            if( AINS_REG_A(iins) != REG_FZERO )
                AinsMakeMove(iins,AINS_REG_A(iins),DT_TFLT,AINS_REG_C(iins),DT_TFLT);
            else
                AinsMakeMove(iins,AINS_REG_B(iins),DT_TFLT,AINS_REG_C(iins),DT_TFLT);
                
        }
        else
        {
            if( AINS_REG_A(iins) != REG_ZERO )
                AinsMakeMove(iins,AINS_REG_A(iins),DT_QUAD,AINS_REG_C(iins),DT_QUAD);
            else
                AinsMakeMove(iins,AINS_REG_B(iins),DT_QUAD,AINS_REG_C(iins),DT_QUAD);
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (ncount) > 0;

    InfoStatus(fp,"\t[OptNormalize%c] Normalized ins: %d#%d {%d ms}\n",CHANGE(change),ncount,ncounth,time);

    return change;
}

/*=======================================================
=======================================================*/
