/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.propagate.c,v 1.24 1998/11/18 22:32:06 muth Exp muth $
  Comment:  This file contains the optimization code for
            Copy Propagation
            Move Elimination
            Reload Avoidance
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

#define AINS_IS_INT_MOVE(iins) (AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0 && AINS_REG_B(iins) != REG_ZERO)
#define AINS_IS_FLT_MOVE(iins) (AINS_CODE(iins) == I_cpys && AINS_REG_A(iins) == AINS_REG_B(iins) && AINS_REG_B(iins) != REG_FZERO)


/*=======================================================
=======================================================*/
#if 0
static REGLIST AinsValueAvailibility(INDEX ains,REG areg, INDEX bins)
{
    INDEX iins;
    REGLIST avail = REG2BIT(areg);

    for(iins = AINS_NEXT(ains); iins!=bins; iins = AINS_NEXT(iins) )
    {
        if( AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0 && (avail & AinsUse(iins)) )
        {
            avail |= AinsDef(iins);
        }
        else if( AINS_CODE(iins) == I_cpys && avail & AinsUse(iins) && AINS_REG_A(iins) == AINS_REG_B(iins) )
        {
            avail |= AinsDef(iins);
        }
        else
        {
            avail &= ~AinsDef(iins);
        }
    }
    
    return avail;
}

#endif

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsValueAvailibilityInExtBbl"

static REGLIST AinsValueAvailibilityInExtBbl(INDEX ains, REGLIST avail, INDEX bins, INDEX ibbl)
{
    INDEX iins;

    if( AINS_BBL(ains) != ibbl )
    {
        CASSERT( AbblNpreds(ibbl) == 1,("weird\n"));
        avail = AinsValueAvailibilityInExtBbl(ains,avail,bins, AEDG_HEAD(ABBL_PRED(ibbl)));
        iins = ABBL_INS_FIRST(ibbl);
    }
    else
    {
        iins = AINS_NEXT(ains);
    }
    
    for( ; VALID(iins) && iins!=bins; iins = AINS_NEXT(iins) )
    {
        if( AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0 && (avail & AinsUse(iins)) )
        {
            avail |= AinsDef(iins);
        }
        else if( AINS_CODE(iins) == I_cpys && avail & AinsUse(iins) && AINS_REG_A(iins) == AINS_REG_B(iins) )
        {
            avail |= AinsDef(iins);
        }
        else
        {
            avail &= ~AinsDef(iins);
        }
    }
    
    return avail;
}




/*=======================================================
@FUN
Return the lowest positon of a set bit in rl.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "RegsGetFirst"

extern REG RegsGetFirst(REGLIST rl)
{
    REG reg = 0;
    
    if( rl == 0 ) return INVALID;
    while( !(rl & 1) )
    {
        rl>>=1;
        reg++;
    }
    return reg;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "DataTypesOk"

static BOOL DataTypesOk(DATATYPE t1, DATATYPE t2)
{
    if( t1 == DT_QUAD && t2== DT_QUAD ) return 1;
    if( t1 == DT_LONG && t2== DT_LONG ) return 1;
    if( t1 == DT_GFLT && t2== DT_GFLT ) return 1;
    if( t1 == DT_TFLT && t2== DT_TFLT ) return 1;
    if( t1 == DT_QUAD && t2 == DT_LONG ) return 1;
    return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindDEfiningLoadStoreInExtBbl"

static INDEX FindDefiningLoadStoreInExtBbl(INDEX load)
{
    INDEX iins;
    INDEX ibbl = AINS_BBL(load);

    iins = AINS_PREV(load);
    
    for(;;)
    {
        for( ; VALID(iins); iins = AINS_PREV(iins) )    
        {
            NUMBER result;

            if( AINS_IS_STORE(iins) )
            {
                result = DoMemoryAccessesConflict(iins,load);
                
                if( result == CONFLICT_MAYBE )
                    return INVALID;
                
                if( result == CONFLICT_NO )
                    continue;
                
                if( result == CONFLICT_MATCH )
                {
                    /* WARNING(("Potential\n")); */
                    return INVALID;
                }
                
                if( result == CONFLICT_EXACT )
                {
                    if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                        return iins;
                    else
                        return INVALID;
                }
                
                ERROR(("unexpected aliasing result store: %d\n",result));
            }
            else if( AINS_TYPE(iins)==IT_ILD || AINS_TYPE(iins) == IT_FLD )
            {
                result = DoMemoryAccessesConflict(iins,load);
                
                if( result == CONFLICT_MAYBE )
                    continue;
                                
                if( result == CONFLICT_MATCH )
                {
                    /* WARNING(("Potential\n")); */
                    continue;
                }
                
                if( result == CONFLICT_NO )
                    continue;
                
                if( result == CONFLICT_EXACT )
                {
                    if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                        return iins;
                    else
                        continue;
                }
                
                ERROR(("unexpected aliasing result load: %d\n",result));
            }
        

        }
        if( AbblNpreds(ibbl) != 1 ) return INVALID;

        ibbl = AEDG_HEAD(ABBL_PRED(ibbl));
        iins = ABBL_INS_LAST(ibbl);
    }
}


/*=======================================================
=======================================================*/                    
#if 0

#undef __PROC__
#define __PROC__ "FindDefiningLoadStore"

static INDEX FindDefiningLoadStore(INDEX load)
{
    INDEX iins;
    
    
    for( iins = AINS_PREV(load); VALID(iins); iins = AINS_PREV(iins) )
    {
        NUMBER result;

        if( AINS_IS_STORE(iins) )
        {
            result = DoMemoryAccessesConflict(iins,load);
            
            if( result == CONFLICT_MAYBE )
                return INVALID;

            if( result == CONFLICT_NO )
                continue;

            if( result == CONFLICT_MATCH )
            {
                /* WARNING(("Potential\n")); */
                return INVALID;
            }
            
            if( result == CONFLICT_EXACT )
            {
                if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                    return iins;
                else
                    return INVALID;
            }
            
            ERROR(("unexpected aliasing result store: %d\n",result));
        }
        else if( AINS_TYPE(iins)==IT_ILD || AINS_TYPE(iins) == IT_FLD )
        {
            result = DoMemoryAccessesConflict(iins,load);
            
            if( result == CONFLICT_MAYBE )
                continue;
            

            if( result == CONFLICT_MATCH )
            {
                /* WARNING(("Potential\n")); */
                continue;
            }
            
            if( result == CONFLICT_NO )
                continue;
            
            if( result == CONFLICT_EXACT )
            {
                
                if( DataTypesOk(AinsDataType(iins),AinsDataType(load)) )
                    return iins;
                else
                    continue;
            }
            

            ERROR(("unexpected aliasing result load: %d\n",result));
        }
        
    }

    return INVALID;
}

#endif
/*=======================================================
@OPT
Avoid reloading of a value from memory.
Pairs of store/load instruction that provably reference the same memory
location are replace by a store/move pair.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptAvoidLoads"

extern BOOL OptAvoidLoads(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Load");
    BOOL   change;

    REGLIST avail;
    
    REG    reg_src;
    REG    reg_dst;

    INDEX load;
    INDEX iins;
    INDEX ifun;
    INDEX ibbl;
    
    NUMBER lfcount = 0;
    NUMBER lfcounth = 0;
    NUMBER lpcount = 0;
    NUMBER lpcounth = 0;

    NUMBER sfcount = 0;
    NUMBER sfcounth = 0;
    NUMBER spcount = 0;
    NUMBER spcounth = 0;
    
    NUMBER bcount = 0;
    NUMBER bcounth = 0;

    INT64 weight = 0;

    clock_t time = clock();

    FOREACH_LIVE_FUN(ifun)
    {
        FOREACH_FUN_BBL(ifun,ibbl)
            FOREACH_BBL_INS(ibbl,load)
            {
                if( AINS_TYPE(load) != IT_ILD && AINS_TYPE(load) != IT_FLD  ) continue;

                iins = FindDefiningLoadStoreInExtBbl(load);
                
                if( AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST )
                {
                    reg_src = AINS_REG_A(iins);

                    if( (AINS_FLAGS(iins) & FI_EVAL_A) && AINS_A_EVAL(iins) != 0 )
                    {
                        POTENTIAL(("Constant forwarding value: %016lx  reg %d  (%ld)\n",
                                 AINS_A_EVAL(iins),reg_src,ABBL_COUNT(AINS_BBL(iins))));
                    }
                    
                    spcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        spcounth++;
                }
                else if( AINS_TYPE(iins) == IT_ILD || AINS_TYPE(iins) == IT_FLD )
                {
                    reg_src = AINS_REG_C(iins);
                    lpcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        lpcounth++;

                }
                else continue;
                

                reg_dst = AINS_REG_C(load);

                if( verbose )
                {
                    fprintf(fp,"\t[*] Load Forwarding Opportunity %d (%d) -> %d (%d):\n",iins,AINS_BBL(iins),load, AINS_BBL(load));
                    fprintf(fp,"\t[*] FROM: ");
                    PrintIns(fp,iins);
                    fprintf(fp,"\t[*] TO  : ");
                    PrintIns(fp,load);
#if 0
                    PrintBblSuper(stdout,AINS_BBL(iins));
                    PrintBblSuper(stdout,ibbl);
#endif
                }


                avail = AinsValueAvailibilityInExtBbl(iins,REG2BIT(reg_src),load,ibbl);

                avail &= REG_IS_INT(reg_dst) ? REGS_INT : REGS_FLT;
                         
                if( avail )
                {
                    AinsMakeMove(load,RegsGetFirst(avail),AinsDataType(iins),reg_dst,AinsDataType(load));
                    goto LabelDone;
                }


                /* we cannot forward it directly, try to use a bypass register */

                avail = MyFindScratchRegsInExtBbl(iins,load);
                                    
                avail &= REG_IS_INT(reg_dst) ? REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                if( avail )
                {
                    INDEX iins_bypass = AinsNew();
                    INDEX jbbl;
                    REG   reg_bypass;

                    reg_bypass = RegsGetFirst(avail);

                    jbbl = ibbl;
                    while( jbbl != AINS_BBL(iins) )
                    {
                        jbbl = AEDG_HEAD(ABBL_PRED(jbbl));
                        ABBL_OUT(jbbl) |= REG2BIT(reg_bypass);
                    }
                    
                               
                    AINS_BBL(iins_bypass) = AINS_BBL(iins);
                    AinsMakeMove(iins_bypass,reg_src,AinsDataType(iins),reg_bypass,AinsDataType(iins));
                    AinsInsertAfter(iins_bypass,iins);
                    AinsMakeMove(load, reg_bypass,AinsDataType(iins),reg_dst,AinsDataType(load));
                    
                    bcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        bcounth++;
                    goto LabelDone;
                }

                if( verbose )
                {
                    fprintf(fp,"\t[*] Failed\n");
                }
                
                continue;

            LabelDone:
                
                if( verbose )
                {
                    fprintf(fp,"\t[*] Load Forwarded %d (%d) -> %d (%d):\n",iins,AINS_BBL(iins),load, AINS_BBL(load));
                    fprintf(fp,"\t[*] FROM: ");
                    PrintIns(fp,iins);
                    fprintf(fp,"\t[*] TO : ");
                    PrintIns(fp,load);
#if 0
                    PrintBblSuper(stdout,AINS_BBL(iins));
                    PrintBblSuper(stdout,ibbl);
#endif
                }

                if( AINS_TYPE(iins) == IT_IST || AINS_TYPE(iins) == IT_FST )
                {
                    sfcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        sfcounth++;
                }
                else
                {
                    lfcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        lfcounth++;
                }

                weight += ABBL_COUNT(AINS_BBL(load));
            }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = sfcount+lfcount>0;

    InfoStatus(fp,"\t[OptLoad%c] SFor %d#%d  SPot: %d#%d LFor: %d#%d  LPot: %d#%d  BPass: %d#%d  wt: %ld {%d ms}\n",
            CHANGE(change),sfcount,sfcounth,spcount,spcounth,lfcount,lfcounth,lpcount,lpcounth,bcount,bcounth,weight,time);
    
    return change;
}

/*=======================================================
@OPT
For each move instruction which moves ra to rb:
Try to replace as many uses of rb with those of rb
This can even handle moves with an offset, eg.lda's
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCopyProp"

extern BOOL OptCopyProp(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Copyprop");
    BOOL   change = 0;
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    INDEX  ibbl;
    INDEX iins;
    INDEX jins;
    REG source;
    REG target;
    
    clock_t time = clock();
    
     
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_CODE(iins) != I_lda && !AINS_IS_FLT_MOVE(iins) )  
                continue;

            source = AINS_REG_B(iins);
            target = AINS_REG_C(iins);

            if( source == target ) continue;
            if( source == REG_ZERO || source == REG_FZERO ) continue;
            if( target == REG_ZERO || target == REG_FZERO ) continue;
            
            for(jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins) )
            {
                if( AINS_REG_A(jins) == target )
                {
                    if( AINS_CODE(iins) == I_cpys )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        AINS_REG_A(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;                                                                            

                    }
                    else if( AINS_CODE(iins) == I_lda && AINS_DATA(iins) == 0)
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }

                        AINS_REG_A(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;                            
                    }
                }

                if( AINS_REG_B(jins) == target )
                {
                    if( AINS_CODE(iins) == I_cpys )
                    {
                        if( verbose )
                        {
                            fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                            PrintBblSuper(fp,ibbl);
                        }

                        AINS_REG_B(jins) = source;
                        pcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            pcounth++;
                    }
                    else if( AINS_CODE(iins) == I_lda)
                    {
                        
                        if( AINS_DATA(iins) == 0)
                        {
                            if( verbose )
                            {
                                fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                                PrintBblSuper(fp,ibbl);
                            }

                            AINS_REG_B(jins) = source;
                            pcount++;
                            if( ABBL_IS_HOT(ibbl) )
                                pcounth++;                            
                        }
                        else if( AINS_CODE(jins) == I_lda ||
                                 AINS_TYPE(jins) == IT_ILD ||
                                 AINS_TYPE(jins) == IT_FLD ||
                                 AINS_TYPE(jins) == IT_ULD ||
                                 AINS_TYPE(jins) == IT_LLD ||
                                 AINS_TYPE(jins) == IT_IST ||
                                 AINS_TYPE(jins) == IT_FST ||
                                 AINS_TYPE(jins) == IT_UST ||
                                 AINS_TYPE(jins) == IT_CST )
                        {
                           if( AINS_DATA(jins) + AINS_DATA(iins) < (1<<15) &&
                               AINS_DATA(jins) + AINS_DATA(iins) >= -(1<<15) )
                           {
                               if( verbose )
                               {
                                   fprintf(fp,"\t[*] modifying instr: %d  move: %d\n",jins,iins);
                                   PrintBblSuper(fp,ibbl);
                               }
                               
                               AINS_REG_B(jins) = source;
                               AINS_DATA(jins) += AINS_DATA(iins);
                               if( AINS_FLAGS(jins) & FI_EVAL_B )
                               {
                                   AINS_B_EVAL(jins) -= AINS_DATA(iins);
                               }
                               pcount++;
                               if( ABBL_IS_HOT(ibbl) )
                                   pcounth++;
                                                                  
                               if( verbose )
                               {
                                   PrintBblSuper(fp,ibbl);
                               }
                           }
                        }
                    }
                }

                if( AINS_REG_C(jins) == target )
                    break;

                if( AINS_REG_C(jins) == source )
                    break;
            }
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = pcount>0;

    InfoStatus(fp,"\t[OptCopyprop%c] Propagated Values: %d#%d {%d ms}\n",CHANGE(change),pcount,pcounth,time);

    return change;
}

/*=======================================================
@OPT
Eliminate Moves

This is a replacement for the copy propagator and the move eliminator
This works only at the bbl level at the moment.

Given a move "move ra,rb" we want to eliminate it!

We proceed as follows

o find the defintion of ra "ra_def" and the last use of ra "ra_last_use"
o the move instruction is the definition of rb "rb_def" find the last
  use of rb "rb_last_use"

  if: rb_last_use AND ra_last_use lie outside of the bbl we cannot eliminate the move! 

  else if: rb_last_use comes before ra_last_use we just rename rb to ra
     this also applied if rb_last_use is outside of the bbl

  else if: ra_last_use comes before rb_last_use we can try to change ra_def to compute
           the values into a different register if rb_last_use is within the bbl
           we cab pick an arbitrary register with the right lifespan.
           if rb_last_use is outside of the bbl we are limited ti rb as our renaming choice.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptMoveElim"

extern BOOL OptMoveElim(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Moveelim");
    BOOL   change = 0;
    
    NUMBER pcount = 0;
    NUMBER pcounth = 0;
    NUMBER kcount = 0;
    NUMBER kcounth = 0;
    INT64  weight = 0;
    clock_t time = clock();

    INDEX  ibbl;
    REG    reg_source;
    REG    reg_target;

    INDEX  iins;
    INDEX  source_def;
    INDEX  target_def;
    INDEX  target_last_use;
    INDEX  source_last_use;
    REGLIST avail;
    REG     bypass;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,target_def)
        {
            if( !AINS_IS_INT_MOVE(target_def) && !AINS_IS_FLT_MOVE(target_def) ) continue;

            reg_source = AINS_REG_B(target_def);
            reg_target = AINS_REG_C(target_def);

            if( reg_target == reg_source ) continue;
            
            source_def = FindInstructionRegDef(target_def, reg_source);
            
                         
            /* source defintion might be outside of the bbl*/

            source_last_use = FindInstructionRegLastUse(target_def, reg_source);
            target_last_use = FindInstructionRegLastUse(target_def, reg_target);

            if( !VALID(target_last_use) && !VALID(source_last_use) ) continue;

            for( iins = target_def; VALID(iins); iins = AINS_NEXT(iins) )
            {
                if( iins == source_last_use || iins == target_last_use )
                {
                    iins = INVALID;
                    break;
                }

                if( AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_FCM ) continue;
                if( AINS_REG_C(iins) == reg_target || AINS_REG_C(iins) == reg_source )
                    break;
            }
            if( VALID(iins) ) continue;
            
            pcount++;
            if( ABBL_IS_HOT(ibbl) )
                pcounth++;

#if 0
            printf("bbl: %d source: def %d  use %d   target: def %d  use %d\n",ibbl,source_def,source_last_use,target_def,target_last_use);
#endif
            /*
              check whether source reg is still available at target_last_use
              this will always be the case if the source range extends to past the target range!
             */
            
            if( VALID(target_last_use) )
            {
                for( iins = target_def; iins != target_last_use; iins = AINS_NEXT(iins) )
                {
                    if( AINS_REG_C(iins) == reg_source )
                    {
                        iins = INVALID;
                        break;
                    }
                    
                }

                if( VALID(iins) )
                {
                    if( verbose )
                    {
                        
                        fprintf(fp,"\t[*] extending the source range over target range for instr: %d\n",target_def);
                        fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                ibbl,source_def,source_last_use,target_def,target_last_use);
                        PrintBblSuper(fp,ibbl);
                    }
                    RenameRegister(target_def,reg_source);
                    AinsKill(target_def);
                    weight += ABBL_COUNT(ibbl);
                    kcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        kcounth++;
                    
                    if( verbose )
                    {
                        PrintBblSuper(fp,ibbl);
                    }
                    continue;
                }
                
            }

            /*
              last use of target before last use of source
              source range does not extend over target range
             */
            
            if( VALID(source_def) && VALID(source_last_use) )
            {
                avail = MyFindScratchRegs(source_def, target_def);
                avail &=  REG_IS_INT(reg_source) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                if( REG2BIT(reg_target) & avail )
                {
                    if( verbose )
                    {
                        fprintf(fp,"\t[*] extending target range over source range  instr: %d\n",target_def);
                        fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                ibbl,source_def,source_last_use,target_def,target_last_use);
                        PrintBblSuper(fp,ibbl);
                    }
                    RenameRegister(source_def,reg_target);
                    AinsKill(target_def);
                    weight += ABBL_COUNT(ibbl);
                    kcount++;
                    if( ABBL_IS_HOT(ibbl) )
                        kcounth++;

                    if( verbose )
                    {
                        PrintBblSuper(fp,ibbl);
                    }
                    
                    continue;
                }
                                

                if( VALID(target_last_use)  )
                {
                    avail = MyFindScratchRegs(source_def, target_last_use);
                    avail &=  REG_IS_INT(reg_source) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

                    if( avail )
                    {
                        bypass = RegsGetFirst(avail);

                        if( verbose )
                        {
                            fprintf(fp,"\t[*] creating new bypass range with reg %d for  instr: %d\n",bypass,target_def);
                            fprintf(fp,"\t[*]bbl: %d source def %d  use %d   target def %d  use %d\n",
                                    ibbl,source_def,source_last_use,target_def,target_last_use);
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        RenameRegister(source_def,bypass);
                        RenameRegister(target_def,bypass);
                        AinsKill(target_def);
                        weight += ABBL_COUNT(ibbl);              
                        kcount++;
                        if( ABBL_IS_HOT(ibbl) )
                            kcounth++;
                        
                        if( verbose )
                        {
                            PrintBblSuper(fp,ibbl);
                        }
                        
                        continue;
                    }
                    
                }
            }
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = kcount>0;

    InfoStatus(fp,"\t[OptMoveelim%c] Eliminated Moves: %d#%d wt: %ld  Potential: %d#%d{%d ms}\n",
            CHANGE(change),kcount,kcounth,weight,pcount,pcounth,time);

    return change;
}

/*=======================================================
=======================================================*/
      
