/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.valprof.c,v 1.1 1998/12/10 20:13:58 muth Exp $
  Comment:  This file contains the code for the optimizations:
                o inlining
                o pathinlining
=======================================================*/
#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"


/* if bbl has been duplicated this contains the index of the clone */
   
#define ABBL_CLONE(i) ABBL_U1(i).doubleword.d1


static INDEX AbblClone(INDEX ibbl)
{
    if( VALID(ABBL_CLONE(ibbl)) ) return ABBL_CLONE(ibbl);
    else return ibbl;
    
}


/*=======================================================
@FUN
Duplicate The Critical Path of a Function
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AfunDupValProf"     

extern void ValProfCloneBbls(INDEX valprof, INT64 value, double flowfrac)
{
    BOOL flag;
    INDEX root;
    INDEX pred;
    INDEX ifun;
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    INDEX jins;
    INDEX iedg;
    REG reg;
    

    CASSERT( value > -(1<<16) && value < (1<<16),("value %016lx out of bounds\n",value));
    CASSERT( valprof < CONFIG_NINS,("weird\n"));
            
    root = AINS_BBL(valprof);
    ifun = ABBL_FUN(root);
    reg = AINS_REG_C(valprof);
    
    ValProfMarkImportantUsesTransitive(valprof);
        
    /* mark bbls that need to be cloned */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_CLONE(ibbl) = INVALID;
        if( ibbl == root ) continue;
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( !(AINS_FLAGS(iins) & F_MARKED) ) continue;
            
            ABBL_CLONE(ibbl) = 1;
        }
        
    }

    
    do
    {
        flag = 0;

        FOREACH_FUN_BBL_R(ifun,ibbl)
        {
            if( ibbl == root ) continue;
            if( ABBL_TYPE(ibbl) == BT_EXIT) continue;
            if( !VALID(ABBL_CLONE(ibbl)) ) continue;
            FOREACH_BBL_PRED(ibbl,iedg)
            {
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
                if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
                pred = AEDG_HEAD(iedg);
                if( pred == root ) continue;
                CASSERT( ABBL_FUN(pred) == ifun,("weird\n"));
                if( VALID(ABBL_CLONE(pred)) ) continue;
                ABBL_CLONE(pred)  = 1;
                flag = 1;
                
            }
        }
    }
    while( flag ); 

    
    /* do everything but the root */
    
    FOREACH_FUN_BBL_R(ifun,ibbl)
    {
        if( ibbl == root ) continue;
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;

        jbbl = AbblDup(ibbl);
        ABBL_FUN(jbbl) = ifun;
        ABBL_CLONE(ibbl) = jbbl;
        ABBL_CLONE(jbbl) = INVALID;
        AbblAppendFancy(jbbl);

        CASSERT( !( ABBL_FLAGS(ibbl) & F_GPREL32),("unimplemented case\n"));
    }


    /* modify  root */
    
    ibbl = AbblNew();    
    jbbl = AbblNew();
    ABBL_TYPE(ibbl) = ABBL_TYPE(root);
    ABBL_TYPE(jbbl) = ABBL_TYPE(root);
    ABBL_FUN(ibbl) = ifun;
    ABBL_FUN(jbbl) = ifun;
    ABBL_COUNT(ibbl) = ABBL_COUNT(root);
    ABBL_CLONE(ibbl) = jbbl;
    ABBL_CLONE(jbbl) = INVALID;

    
    jins = AinsNew();
    AinsMakeLda(jins,reg, reg, value);
    AINS_BBL(jins) = ibbl;
    AinsAppend(jins);

    jins = AinsNew();
    AinsMakeLda(jins,reg, reg, value);
    AINS_BBL(jins) = jbbl;
    AinsAppend(jins);
        
    for( iins = AINS_NEXT(valprof); VALID(iins); iins = AINS_NEXT(iins) )
    {
        jins = AinsDup(iins);
        AINS_BBL(jins) = ibbl;
        AinsAppend(jins);

        jins = AinsDup(iins);
        AINS_BBL(jins) = jbbl;
        AinsAppend(jins);

        AinsKill(iins);
    }

    AbblAppendFancy(jbbl);
    AbblAppendFancy(ibbl);
                  
    CERROR( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL,("unimplemented\n"));

    FOREACH_BBL_SUCC(root,iedg)
    {
        if( AEDG_IS_INTER(iedg) )
        {
            ERROR(("unexpected inter edge\n"));
            PrintEdgSuper(stdout,iedg);
        }

        AedgLinkFancy( AedgNewFancy(ibbl,AEDG_TAIL(iedg),AEDG_TYPE(iedg)));
        AedgKillFancy(iedg);
    }
            
    jins = AinsNew();
    AinsMakeLda(jins,reg, reg, -value);
    AINS_BBL(jins) = root;
    AinsAppend(jins);

    jins = AinsNew();
    AinsMakeBne(jins,reg, 0);
    AINS_BBL(jins) = root;
    AinsAppend(jins);
    ABBL_TYPE(root) = BT_BRANCH;
    
    AedgLinkFancy( AedgNewFancy(root,ibbl,ET_TRUE));
    AedgLinkFancy( AedgNewFancy(root,jbbl,ET_FALSE));

    /* adapt counts */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;
        ABBL_COUNT(ABBL_CLONE(ibbl)) = (double) ABBL_COUNT(ibbl) * flowfrac;
        ABBL_COUNT(ibbl) = (double) ABBL_COUNT(ibbl) * (1.0 - flowfrac);
    }
        
    /* duplicate edges leaving clones */
    
    FOREACH_FUN_BBL_R(ifun,ibbl)
    {
        if( ibbl == root ) continue;
        if( ABBL_TYPE(ibbl) == BT_EXIT) continue;
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;

        /* invariant: link edge before call edge */

        if( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL )
        {
            iedg = ABBL_EDG_LINK(ibbl);            
            AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AbblClone(AEDG_TAIL(iedg)),AEDG_TYPE(iedg)) );
            iedg = ABBL_EDG_CALL(ibbl);
            AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AEDG_TAIL(iedg),AEDG_TYPE(iedg)) );
        }
        else
        {
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) )
                {
                    ERROR(("unexpected inter edge\n"));
                    PrintEdgSuper(stdout,iedg);
                }
                AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AbblClone(AEDG_TAIL(iedg)),AEDG_TYPE(iedg)) );
            }
        }

    }

}

/*=======================================================
=======================================================*/





