/*=======================================================
  Author:   R. Muth
  Version:  $Id: optimize.c,v 1.72 1998/12/10 00:14:11 muth Exp $
  Comment:  This file contains the optimization code
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

#define CUTOFF_PERCENTILE_HARD 0.33
#define CUTOFF_NINS_HARD 0

/*=======================================================
@FUNCTION
Iteratively Execute Easy Optimization
This function calls all the easy optimizations.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeEasy"

extern BOOL OptimizeEasy(FILE *fp)    
{
    BOOL flag = 0;
    BOOL cont;

#if 0
    static BOOL x =0;

    if( x )
        return 0;
    else
        x = 1;
#endif

    InfoPhase(fp,"Optimize Simple");    
    
    /***************************************************/
    
    ComputeCountCutOff(fp,100,0.66,1500);

    InfStatResource(fp);
    
    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prenoop") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Noop") )
        flag |= OptEliminateNoops(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postnoop") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

    if( HasOption(PRINT_OPT_WHEN,"preReach") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Reach") )
    {
        flag |= OptRemoveUnreachableCode(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postReach") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    InfStatResource(fp);
        
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"preNormalize") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Normalize") )
    {
        flag |= OptNormalize(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postNormalize") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"preCalleesave") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Calleesave") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptCalleeSaveRegisters(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postCalleesave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"preDead") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Dead") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
             flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postDead") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prepeep") )
        PrintAll(stdout);
        
    if( !HasOption(DISABLED_OPT,"Peep") )
        flag |= OptPeepHole(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postpeep") )
    {
        /* needs liveness */
        PrintAll(stdout);
    }
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preliteral") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Literal") )
    {
        /* needs liveness */
        flag |= OptLiteral(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postliteral") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preeval") )
        PrintAll(stdout);
           
    if( !HasOption(DISABLED_OPT,"Eval") )
        flag |= OptPartialEvaluate(fp);
       
    if( HasOption(PRINT_OPT_WHEN,"Posteval") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    /* this should go before reachabilty analysis to avoid
       some inconsistencies (fall thru convention) */

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preforward") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Forward") )
        flag |= OptBranchForwarding(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postforward") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Precond") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Cond") )
    {
        flag |= OptBranchConditional(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcond") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preempty") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Emptybbl") )
    {
        flag |= OptUnlinkEmptyBbls(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postempty") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Premerge") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Mergebbl") )
        flag |= OptBranchMerge(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postmerge") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preconstone") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Constone") )
        flag |= OptConstantsOne(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postconstone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preconsttwo") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Consttwo") )
    {
        flag |= OptConstantsTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postconsttwo") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preloop") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Loop") )
    {
        OptRemoveUnreachableCode(fp);
        ComDomsAndLoops(fp);
        ComStackInfo(fp);
        ComputeAliasAuxiliaryInformation(fp);
            
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComputeFunctionSideEffects(fp);
 
        flag |= OptMoveLoopInvariantCode();
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postloop") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

#endif

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preload") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Load") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComputeAliasAuxiliaryInformation(fp);
        ComLiveness(fp);
        
        flag |= OptAvoidLoads(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postload") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preswitch") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Switch") )
        flag |= OptTableJumps(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postswitch") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Pregprel") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Gprel") )
        flag |= OptGprel(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postgprel") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Precompare") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Compare") )
     {
         ComputeSavedChangedRegisters(fp);
         ComLiveness(fp);
         flag |= OptCompare(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postcompare") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
 
    /***************************************************/
#if 0
     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prestacksave") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Stacksave") )
    {
        ComputeSavedChangedRegisters(fp);
        flag |= OptStackSaveRegs(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststacksave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif
    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prestackdel") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Stackdel") )
    {
        ComStackInfo(fp);
        flag |= OptDeleteStack(fp);
        flag |= OptKillDeadStores(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststackdel") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prespec") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Spec") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComputeAliasAuxiliaryInformation(fp);
            
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);

        flag |= OptUndoSpeculations(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postspec") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/
#if 0
    if( HasOption(PRINT_OPT_WHEN,"Prepleafsave") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Leafsave") )
    {
        ComputeSavedChangedRegisters(fp);
        ComputeLiveness(fp);
        flag |= OptLeafSaveRegisters(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"PostLeafsave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Premove") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Move") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptMoveElim(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmove") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Precopy") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Copy") )
    {
        flag |= OptCopyProp(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcopy") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
    /* I think it is important to cleanup after the copy
       propagation because otherwise the next optimization
       might undo some of the effects. */
       
    if( HasOption(PRINT_OPT_WHEN,"preDead") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Dead") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptKillUseless(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postDead") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preescapes") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Escapes") )
    {
        flag |= OptSpecialEliminateEscapes(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postescapes") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
    /* this should happen before the conditional move stuff */

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Premergeinsup") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Mergeinsup") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptMergeInsUp(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmergeinsup") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Precondpropagate") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condpropagate") )
    {
        flag |= OptCondPropagate(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondpropagate") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Precondmoveone") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condmoveone") )
    {
        flag |= OptCondMoveOne(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondmoveone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

    if( HasOption(PRINT_OPT_WHEN,"Precondmovetwo") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condmovetwo") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptCondMoveTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondmovetwo") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Predupbblone") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Dupbblone") )
    {
        flag |= OptDupBblOne(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbblone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preinlinesmall") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Inlinesmall") )
     {
         flag |= OptInlineSmall(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postinlinesmall") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
 
     /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
     
    if( HasOption(PRINT_OPT_WHEN,"Predupbbltwo") )
         PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Dupbbltwo") )
    {
        flag |= OptDupBblTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbbltwo") )
        PrintAll(stdout);
 
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    
    
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
     
    if( HasOption(PRINT_OPT_WHEN,"Predupbblthree") )
         PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Dupbblthree") )
    {
        flag |= OptDupBblThree(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbblthree") )
        PrintAll(stdout);
 
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
     /***************************************************/    

     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
         
     if( HasOption(PRINT_OPT_WHEN,"Preregreallocone") )
         PrintAll(stdout);
      
     if( !HasOption(DISABLED_OPT,"Regreallocone") )
     {
          ComputeSavedChangedRegisters(fp);
          ComLiveness(fp);
          OptRegisterReallocOne(fp);
      }
     
     if( HasOption(PRINT_OPT_WHEN,"Postregreallocone") )
         PrintAll(stdout);
     
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
     
      /***************************************************/

     InfStatResource(fp);

     return flag;    
}


/*=======================================================
@FUNCTION
The Hard Optimization.
This function is called only once sandwiched between calls
to OptimizeEasy.
It calls all the optimizations that should be applied only once.
Because of problems with jumptable an ugly call to OptimizeEasy
was necessary.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeHard"


extern BOOL OptimizeHard(FILE *fp)    
{

    BOOL flag = 0;

    InfoPhase(fp,"Optimize Hard");    
    
    if( HasOption(DISABLED_OPT,"Hard") )
        return 0;

    /***************************************************/
    
    ComputeCountCutOff(fp,100,CUTOFF_PERCENTILE_HARD,CUTOFF_NINS_HARD);        

    InfStatResource(fp);

#if 0
    ComUseDef(stdout,1);

    ValProfCloneBbls(15344, 0, 1.00);

#endif
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHEN,"Precallingconvention") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Callingconvention") )
    {
        OptCallingConvention(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postcallingconvention") )
        PrintAll(stdout);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHEN,"Preshrinkwrap") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Shrinkwrap") )
    {
      ComDomsAndLoops(fp);
      OptShrinkWrap(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postshrinkwrap") )
        PrintAll(stdout);

    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHEN,"Predupfun") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Dupfun") )
    {
        OptRemoveUnreachableCode(fp);
        flag |= OptDuplicateFunctions(fp);
    }

    
    if( HasOption(PRINT_OPT_WHEN,"Postdupfun") )
        PrintAll(stdout);
#endif
    /***************************************************/
#if 1

    OptimizeEasy(fp);
                                 
#endif    
    /***************************************************/    

    ComputeCountCutOff(fp,100,CUTOFF_PERCENTILE_HARD,CUTOFF_NINS_HARD);

    InfStatResource(fp);

    /***************************************************/
        
    if( HasOption(PRINT_OPT_WHEN,"Preinline") )
        PrintAll(stdout);

    IndirInit();

    /***************************************************/    

    if( !HasOption(DISABLED_OPT,"Indir") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComDomsAndLoops(fp);

	OptInlineIndir(fp);
    }


    if( HasOption(PRINT_OPT_WHEN,"PostIndir") ) 
        PrintAll(stdout);

    /***************************************************/    
#if 0
    if( !HasOption(DISABLED_OPT,"Jump") )
    {
        ComputeNumberOps(fp);

        OptInlineJumps(fp);

        ComputeNumberOps(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"PostJump") ) 
        PrintAll(stdout);
#endif
    /***************************************************/
   

    IndirExit();
    
    if (!HasOption(DISABLED_OPT,"Inline") )
    {
	if( HasOption(INLINE_METHOD,"Crit") )
	{
	    ComputeCriticalBbls(fp);
	    ComStackInfo(fp);
	    OptInline(fp,0);
	}
	else if( HasOption(INLINE_METHOD,"Path") )
	{
	    ComputeCriticalBbls(fp);
	    ComStackInfo(fp);
	    OptInline(fp,1);
	}
    }

#if 0
    ComputeFunctionSideEffects(stdout);
    ComputeMod2kptrs();
#endif
    
    if( HasOption(PRINT_OPT_WHEN,"Postinline") )
        PrintAll(stdout);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preinlineonecall") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Inlineonecall") )
     {
         flag |= OptInlineOneCall(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postinlineonecall") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
 
     fflush(fp);
     

#if 1
    /***************************************************/

    if( HasOption(PRINT_OPT_WHEN,"Prestackmerge") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Stackmerge") )
    {
        ComStackInfo(fp);
        flag |= OptStackMerge(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststackmerge") )
        PrintAll(stdout);

    /***************************************************/    
#endif

    return flag;
}


/*=======================================================
=======================================================*/


