/*=======================================================
  Author:   
  Version:  $Id$
  Comment:  
=======================================================*/


#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
/*=======================================================
Special treatment for setjmp/longjmp style function.
Mark those functions and add special edges to get
the controlflow right.
=======================================================*/
#undef __PROC__
#define __PROC__ "SpecialSetJmpLongJmp"

extern void SpecialSetJmpLongJmp(FILE *fp)
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Setjmp");
    INDEX iins;
    NUMBER count = 0;
    INDEX ifun;
    
    INDEX jins;
    
    FOREACH_LIVE_INS(iins)
    {
        /* heuristic identifying setjmp style functions */
        
        if( AINS_CODE(iins) == I_stq &&
            AINS_REG_A(iins) == REG_RA &&
            AINS_REG_B(iins) != REG_SP &&
            AINS_REG_B(iins) != REG_FP &&
            AinsDataType(iins) == DT_QUAD)
        {
            for( jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins) )
            {
                if( AINS_REG_C(jins) == REG_RA ) break;
                
                if( AINS_CODE(jins) == I_ret )
                {
                    ifun = ABBL_FUN(AINS_BBL(iins));
                    if( verbose )
                    {
                        InfoExtra(fp,"\t[*] Compensating PC store in fun: %d [%s]\n",ifun,AFUN_NAME(ifun));
                    }
#if 0
                    PrintIns(stdout,iins);
#endif
                    count++;
                    AedgLinkFancy( AedgNewFancy(AFUN_BBL_LAST(AFUN_PSEUDO_HELL),
                                                AFUN_BBL_LAST(ifun),
                                                ET_SETLONGJMP));
                    AFUN_FLAGS(ifun) |= FF_SETJMP;
                    break;
                }
            }
            
            continue;
        }

        /* heuristic identifying slongjmp style functions */
        
        if( AINS_CODE(iins) == I_ldq &&
            AINS_REG_C(iins) == REG_RA &&
            AINS_REG_B(iins) != REG_SP &&
            AINS_REG_B(iins) != REG_FP &&
            AinsDataType(iins) == DT_QUAD)
        {
            for( jins = AINS_NEXT(iins); VALID(jins); jins = AINS_NEXT(jins) )
            {
                if( AINS_REG_C(jins) == REG_RA ) break;

                ifun = ABBL_FUN(AINS_BBL(iins));
                if( AINS_CODE(jins) == I_ret )
                {
                    if( verbose )
                    {
                        InfoExtra(fp,"\t[*] Compensating PC load in fun: %d [%s]\n",ifun,AFUN_NAME(ifun));
                    }
#if 0
                    PrintIns(stdout,iins);
#endif
                    count++;
                    AedgLinkFancy( AedgNewFancy(AFUN_BBL_LAST(ifun),
                                                AFUN_BBL_LAST(AFUN_PSEUDO_HELL),
                                                ET_SETLONGJMP));
                    AFUN_FLAGS(ifun) |= FF_LONGJMP;
                    break;
                }
            }
            
            continue;
        }   
    }
    
    InfoStatus(fp,"\t[SpecialSetjmp] compensations added: %d\n",count);    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SpecialEliminateHeadFunctions"

extern void SpecialEliminateHeadFunctions(FILE *fp)
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Headfun");
    INDEX iedg;
    INDEX pred;
    INDEX fcount = 0;
    
#if 1
    FOREACH_LIVE_EDG(iedg)
    {
        INDEX hd = AEDG_HEAD(iedg);
        INDEX tl = AEDG_TAIL(iedg);
        INDEX ifun = ABBL_FUN(hd);
        
        if( !AEDG_IS_INTER(iedg) ) continue;
        if( AEDG_TYPE(iedg) != ET_NORM ) continue;
        
        if( ABBL_HAS_INSS(hd) ) continue;
        if( !ABBL_IS_FIRST(hd) ) continue;
        if( !ABBL_IS_FIRST(tl) ) continue;

        fcount++;
        
        if( verbose )
        {
            InfoExtra(fp,"\t[*] deleting head fun: %d [%s]\n",ifun,AFUN_NAME(ifun));
        }
        
        AedgKillFancy(iedg);
                
        FOREACH_BBL_PRED(hd,pred)
        {
                INDEX jedg = AedgNewFancy(AEDG_HEAD(pred),tl,AEDG_TYPE(pred));
                AedgKillFancy(pred);
                AedgLinkFancy(jedg);
        }
        
        /* return code should be checked here !!!!*/

        AbblMigrateRelocInfo(hd,tl);
                
        AbblKill(hd);
        AfunKill(ABBL_FUN(hd));
    }
    
    InfoStatus(fp,"\t[SpecialHeadfun] funs: %d\n",fcount);    
#endif 
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComVerifyGp"

static void ComVerifyGp(FILE *fp)
{
    INDEX iins;
    INT64 curr = 0;
    INT64 gp = SysGetGlobalPointer();
    
    FOREACH_INS(iins)
    {
        if( AINS_IS_KILLED(iins) ) continue;
        if( AINS_FLAGS(iins) & F_REFQUAD )
            curr = Ins2Addr(iins);

        if( AINS_CODE(iins) == I_bsr || AINS_CODE(iins) == I_jsr )
            curr = Ins2Addr(iins+1);

        if( AINS_CODE(iins) == I_br && AINS_REG_C(iins) != REG_ZERO )
            curr = Ins2Addr(iins+1);
        
        if( AINS_REG_C(iins) != REG_GP ) continue;
        if( AINS_TYPE(iins) != IT_LDA ) continue;

        if( AINS_REG_C(iins) != REG_GP ) continue;

        if( AINS_CODE(iins) == I_lda )
            curr += AINS_DATA(iins);
        else if( AINS_CODE(iins) == I_ldah )
            curr += AINS_DATA(iins) << 16;
        else
            ERROR(("weird\n"));


        if( AINS_REG_B(iins) == REG_GP && AINS_REG_C(iins+1) != REG_GP )
        {
            if( curr != gp )
            {
#if 0
                WARNING(("potentially bad gp %016lx\n",curr));
                PrintBblSuper(stdout,AINS_BBL(iins));
#endif		
            }
            
        }
    }

    InfoStatus(fp,"\t[ComVerifygp] gp is (%lx)\n",gp);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SpecialGlobalPointers"

extern BOOL SpecialGlobalPointers(FILE *fp)
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Gp");
    BOOL change;
    INDEX ifun;
    INDEX ibbl;
    INDEX iins;
    INDEX startIns;
    INDEX killDistance;
    INDEX entry;
    NUMBER bcount = 0;
    NUMBER lcount = 0;
    NUMBER mcount = 0;
    ADDRESS Y;
    NUMBER high, low;
    

    ComVerifyGp(stdout);

    entry = Addr2Ins(SysGetProgramEntry());

#ifdef SYS_COFF
    killDistance = 10;
#elif SYS_ELF
    killDistance = 5;
#endif


    FOREACH_LIVE_INS(iins)
    {
        ibbl = AINS_BBL(iins);
        ifun = ABBL_FUN(ibbl);

#if 0
         /* NEEDS WORK: do not kill the initial gp setting */
	if ((iins > entry) && (iins < entry + killDistance))
	    continue;
#endif
	

        if( AINS_CODE(iins) == I_br && AINS_REG_C(iins) != REG_ZERO )
        {
            INDEX iedg = ABBL_EDG_EXTRA(ibbl);
            INDEX jedg;
                
            CASSERT( AEDG_TYPE(iedg) == ET_EXTRA,("Weird\n"));
            CASSERT( AEDG_TAIL(iedg) == ABBL_NEXT(ibbl),("Weird\n"));
            jedg = AedgNewFancy(AEDG_HEAD(iedg),AEDG_TAIL(iedg),ET_NORM);
            AedgKillFancy(iedg);
            AinsKill(iins);
            ABBL_TYPE(ibbl) = BT_NORM;
            AedgLinkFancy(jedg);
                
            if( verbose )
            {
                InfoExtra(fp,"\t[*] Killing br in bbl %d fun %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun));
                PrintIns(fp,iins);
            }
            
            bcount++;


            
        }
        else if( AINS_REG_C(iins) == REG_GP )
        {
            if( AINS_CODE(iins) == I_lda || AINS_CODE(iins) == I_ldah )
            {
                AinsKill(iins);
                lcount++;
            }
            else
            {
                if( verbose )
                {
                    InfoExtra(fp,"\t[*] Killing strange iins in bbl %d fun %d [%s]\n",
                            ibbl,ifun,AFUN_NAME(ifun));
                    PrintIns(fp,iins);
                }
                
                AinsKill(iins);
                mcount++;
            }
        }
    }


    /* set the global pointer as the first 3-4 instructions */
    startIns = Addr2Ins(SysGetProgramEntry());
    ibbl = AINS_BBL(startIns);
    
    iins = AinsNew();
    Disassemble(0xc3a00000, &AINS(iins));
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, startIns);

    /* now add a bunch of stuff to gp to set it correctly */
    Y = SysGetGlobalPointer() - (SysGetProgramEntry() + 4);

    high = Y >> 16;
    low = Y & 0xffff;
    if( low & 0x8000 )
    {
	low |= 0xffff0000;
	high += 1;
    }

#if 0
    WARNING(("Y %16lx  high %8lx  low %8x\n",Y,high,low));
#endif
    CASSERT( ((long)high << 16) + (long) low == Y,("bad table address computation\n"));
    
    
    iins = AinsNew();
    AinsMakeLdah(iins, REG_GP, REG_GP, high);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, startIns);

    iins = AinsNew();
    AinsMakeLda(iins, REG_GP, REG_GP, low);
    AINS_BBL(iins) = ibbl;
    AinsInsertBefore(iins, startIns);
    

    change = (bcount+lcount+mcount > 0);

    InfoStatus(fp,"\t[OptGp%c] GP: %016lx   lda(h): %d  br: %d  misc: %d\n",
           CHANGE(change),SysGetGlobalPointer(),lcount,bcount,mcount);

    return change;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "OptSpecialEliminateEscapes"

extern BOOL OptSpecialEliminateEscapes(FILE *fp)
{
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Escapes");
    BOOL change;
    INDEX iedg;
    INDEX ecount = 0;
    INDEX pcount = 0;
    
    FOREACH_LIVE_EDG(iedg)
    {
        INDEX hd = AEDG_HEAD(iedg);
        INDEX tl = AEDG_TAIL(iedg);

        if( AFUN_IS_PSEUDO(ABBL_FUN(hd)) ) continue;
        if( AFUN_IS_PSEUDO(ABBL_FUN(tl)) ) continue;
        
        if( ABBL_FUN(hd) == ABBL_FUN(tl) ) continue;
        if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
        if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;        

        pcount++;

        if( VALID(AFUN_MASTER(ABBL_FUN(hd))) ) continue;

                                               
        if( ABBL_TYPE(tl) == BT_RET )
        {
            INDEX ifun = ABBL_FUN(hd); 

            INDEX ibbl;

            if( verbose )
            {
                fprintf(fp,"\t[*] eliminating esc to bt_ret in fun %d [%s]\n",ifun,AFUN_NAME(ifun));
                PrintBblSuper(stdout,hd);
                PrintBblSuper(stdout,tl);
                
            }


            
            ibbl = AbblDup(tl);   /* Count field is not correct: NEEDS WORK */
            ABBL_FUN(ibbl) = ifun;

            AbblAppendFancy(ibbl);

            AedgKillFancy(iedg);
            AedgLinkFancy( AedgNewFancy(hd,ibbl,AEDG_TYPE(iedg)));
            AedgLinkFancy( AedgNewFancy(ibbl,ibbl,ET_EXIT_LINK));

            ecount++;

        }
        else if( ABBL_TYPE(tl) == BT_PAL )
        {
            INDEX ibbl_ret = AEDG_TAIL(ABBL_EDG_LINK(tl));

            if( ABBL_TYPE(ibbl_ret) == BT_RET )
            {
                INDEX ifun;
                INDEX ibbl;
                INDEX jbbl;
                
                ibbl = AbblDup(tl);   /* Count field is not correct: NEEDS WORK */
                ifun = ABBL_FUN(hd); 
                ABBL_FUN(ibbl) = ifun;   
                AbblAppendFancy(ibbl);
                jbbl = AbblDup(ibbl_ret);   /* Count field is not correct: NEEDS WORK */
                ABBL_FUN(jbbl) = ifun;   
                AbblAppendFancy(jbbl);

                AedgKillFancy(iedg);
                AedgLinkFancy( AedgNewFancy(hd,ibbl,AEDG_TYPE(iedg)));
                AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_FUN_LINK));
                AedgLinkFancy( AedgNewFancy(ibbl,AFUN_BBL_FIRST(AFUN_PSEUDO_PAL),ET_FUN_CALL));
                AedgLinkFancy( AedgNewFancy(jbbl,jbbl,ET_EXIT_LINK));

                ecount++;
                if( verbose )
                {
                    InfoExtra(fp,"\t[*] eliminating esc to bt_pal in fun %d [%s]\n",ifun,AFUN_NAME(ifun));
                }
            }
        }
    }

    change = (ecount>0);

    InfoStatus(fp,"\t[OptSpecialEliminateEscapes%c] edgs: %d/%d\n",CHANGE(change),ecount,pcount);    
    return change;
}

/*=======================================================
=======================================================*/
