/*=======================================================
  Author:   R. Muth
  Version:  $Id: system.coff.c,v 1.6 1998/12/08 20:13:41 muth Exp muth $
  Comment:  This file contains platform specific functions
=======================================================*/

#include "global.h"
#include "functions.h"
#include "system.coff.h"
#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
extern INT64 SysGetGlobalPointer()
{
    return BIN_GP;
}
/*=======================================================
=======================================================*/
extern ADDRESS SysGetProgramEntry()
{
    return BIN_ENTRY;
}

/*=======================================================
=======================================================*/
extern INDEX SysGetTextSection()
{
    return BIN_SECTION_MAP[R_SN_TEXT];
}
/*=======================================================
=======================================================*/
extern INDEX SysGetSwitchSection()
{
    return BIN_SECTION_MAP[R_SN_DATA];
}
/*=======================================================
=======================================================*/
extern void SysUpdateTextSegmentSize()
{
    /*
      we assume that the text section is the last section within the text segment
      without this assumption we could not even enlarge the text section
     */
    INDEX isec = SysGetTextSection();
#if 0
    printf("Old: TextSegemnt size is: %016lx\n",BIN_AOUTHDR.tsize);
#endif
    BIN_AOUTHDR.tsize = ASEC_START(isec) + ASEC_SIZE(isec) - BIN_AOUTHDR.text_start;
    BIN_AOUTHDR.tsize = (BIN_AOUTHDR.tsize + 0x2000) & ~(0x2000-1);
#if 0
    printf("New: TextSegemnt size is: %016lx\n",BIN_AOUTHDR.tsize);
#endif    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "Addr2Ins"

extern INDEX Addr2Ins(ADDRESS address)
{
    INDEX isec_text = BIN_SECTION_MAP[R_SN_TEXT];
    return (address-ASEC_START(isec_text))>>2;    
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "Ins2Addr"

extern ADDRESS Ins2Addr(INDEX iins)
{
    INDEX isec_text = BIN_SECTION_MAP[R_SN_TEXT];
    if( !VALID(iins) )
        return INVALID;
    return (iins*4) + ASEC_START( isec_text );
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindSymbol"

extern STRING FindSymbol(ADDRESS address)
{
    UINT32 is;

    FOREACH_ESY(is)
    {
        if( AESY_VALUE(is) == address )
            return AESY_NAME(is);
    }

    FOREACH_LSY(is)
    {
        if( ALSY_VALUE(is) == address )
            return ALSY_NAME(is);
    }

    address -=8;
    
    FOREACH_ESY(is)
    {
        if( AESY_VALUE(is) == address )
            return AESY_NAME(is);
    }

    FOREACH_LSY(is)
    {
        if( ALSY_VALUE(is) == address )
            return ALSY_NAME(is);
    }

    
    return "++undef++";
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "IsFunctionStart"

extern BOOL IsFunctionStart(ADDRESS address)
{
    UINT32 is;

    FOREACH_ESY(is)
    {
        if( AESY_VALUE(is) == address )
            return AESY_IS_PROC(is);
    }

    FOREACH_LSY(is)
    {
        if( ALSY_VALUE(is) == address )
            return ALSY_IS_PROC(is);
    }

    address -=8;
    
    FOREACH_ESY(is)
    {
        if( AESY_VALUE(is) == address )
            return AESY_IS_PROC(is);
    }

    FOREACH_LSY(is)
    {
        if( ALSY_VALUE(is) == address )
            return ALSY_IS_PROC(is);
    }

    
    return 1;
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "LocateSection"

/* we allow the address to point to the end of a section - dangerous */

extern INDEX LocateSection(ADDRESS address)
{
    UINT32 is;
    
    FOREACH_SEC(is)
    {
        if( AsecContainsAddress(is,address) )
            return is;
    }
    
    is = BIN_SECTION_MAP[R_SN_TEXT];
    if( address==ASEC_START(is)+ASEC_SIZE(is) )
    {
#if 0        
        WARNING(("access to end of text section \n"));
#endif        
        return is;
    }
    else
    {
#if 0        
        WARNING(("Cannot locate address: %lx\n",address));
#endif        
    }
    
    return INVALID;
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecNew"

extern INDEX AsecNew()
{
    CFATAL(CONFIG_USEC >= CONFIG_NSEC,("not enough sections\n"));
    return(CONFIG_USEC++);
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecSetSize"

extern void AsecSetSize(INDEX isec, INT64 size)
{
    ASEC_SIZE(isec) = size;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecGetSize"

extern INT64 AsecGetSize(INDEX isec)
{
    return ASEC_SIZE(isec);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecContainsAddress"

extern BOOL AsecContainsAddress(INDEX isec,ADDRESS addr)
{
    if( addr >= ASEC_START(isec) && addr < ASEC_START(isec) + ASEC_SIZE(isec) )
        return 1;
    else
        return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AddressIsCode"

extern BOOL AddressIsCode(ADDRESS addr)
{
    INDEX isec;

    isec = BIN_SECTION_MAP[R_SN_TEXT];
    if( addr >= ASEC_START(isec) && addr < ASEC_START(isec) + ASEC_SIZE(isec) )
        return 1;

    isec = BIN_SECTION_MAP[R_SN_INIT];
    if( addr >= ASEC_START(isec) && addr < ASEC_START(isec) + ASEC_SIZE(isec) )
        return 1;

    isec = BIN_SECTION_MAP[R_SN_FINI];
    if( addr >= ASEC_START(isec) && addr < ASEC_START(isec) + ASEC_SIZE(isec) )
        return 1;
    
    return 0;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecMapAddress"

extern ADDRESS AsecMapAddress(INDEX isec,ADDRESS addr)
{
    CASSERT( AsecContainsAddress(isec,addr),("weird\n"));
    return addr - ASEC_START(isec) + (INT64)ASEC_DATA8(isec);
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecIsReadOnly"

extern BOOL AsecIsReadOnly(INDEX isec)
{
    if( ASEC_IS_RDATA(isec)  ||
        ASEC_IS_TEXT(isec)   ||
        ASEC_IS_INIT(isec)   ||
        ASEC_IS_FINI(isec)   ||
        ASEC_IS_LIT8(isec)   ||
        ASEC_IS_LIT4(isec)   ||
        ASEC_IS_LITA(isec)   ||
        ASEC_IS_RCONST(isec) )
        return 1;
    else
        return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecWriteData32"

extern void AsecWriteData32(INDEX isec, NUMBER index,INT32 value)
{
    /*
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    */
    ASEC_DATA32(isec)[index] = value;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecReadData32"

extern INT32 AsecReadData32(INDEX isec, NUMBER index)
{
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    return ASEC_DATA32(isec)[index];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecGetAddressData32"

extern ADDRESS AsecGetAddressData32(INDEX isec, NUMBER index)
{
    CASSERT( index*4 < ASEC_SIZE(isec),("weird\n"));
    return ASEC_START(isec) + 4 *index;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AsecIsCode"

extern BOOL AsecIsCode(INDEX isec)
{
    if( ASEC_IS_TEXT(isec)   ||
        ASEC_IS_INIT(isec)   ||
        ASEC_IS_FINI(isec)   )
        return 1;
    else
        return 0;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData64"

extern INT64 ReadData64(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA64(is),("illegal access\n"));
    return ASEC_DATA64(is)[(address-ASEC_START(is))>>3];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "WriteData64"

extern void WriteData64(ADDRESS address, INT64 value)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA64(is),("illegal access\n"));
    ASEC_DATA64(is)[(address-ASEC_START(is))>>3]  = value ;
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData32"

extern INT32 ReadData32(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA32(is),("illegal access\n"));
    return ASEC_DATA32(is)[(address-ASEC_START(is))>>2];
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "WriteData32"

extern void WriteData32(ADDRESS address, INT32 value)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA32(is),("illegal access\n"));
    ASEC_DATA32(is)[(address-ASEC_START(is))>>2] = value;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ReadData16"

extern INT16 ReadData16(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA16(is),("illegal access\n"));
    return ASEC_DATA16(is)[(address-ASEC_START(is))>>1];
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FetchData8"

extern INT8 ReadData8(ADDRESS address)
{
    UINT32 is = LocateSection(address);
    CASSERT(ASEC_DATA8(is),("illegal access\n"));
    return ASEC_DATA8(is)[(address-ASEC_START(is))];
}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintRELOC"

static char *sn[] =
{
    "R_SN_NULL  ",
    "R_SN_TEXT  ",
    "R_SN_RDATA ",
    "R_SN_DATA  ",
    "R_SN_SDATA ",
    "R_SN_SBSS  ",
    "R_SN_INIT  ",
    "R_SN_LIT8  ",
    "R_SN_LIT4  ",
    "R_SN_XDATA ",
    "R_SN_PDATA ",
    "R_SN_FINI  ",
    "R_SN_LITA  ",
    "R_SN_ABS   ",
    "R_SN_RCONST"
};

static char *type[] =
{
    "R_ABS       ",
    "R_REFLONG   ",
    "R_REFQUAD   ",
    "R_GPREL32   ",
    "R_LITERAL   ",
    "R_LITUSE    ",
    "R_GPDISP    ",
    "R_BRADDR    ",
    "R_HINT      ",
    "R_SREL16    ",
    "R_SREL32    ",
    "R_SREL64    ",
    "R_OP_PUSH   ",
    "R_OP_STORE  ",
    "R_OP_PSUB   ",
    "R_OP_PRSHIFT",
    "R_GPVALUE   ",
    "R_GPRELHIGH ",
    "R_GPRELLOW  ",
    "R_IMMED     ",
};



static void PrintRELOC(FILE *fp,RELOC *r)
{
    fprintf(fp,
            "r_vaddr:    %016lx\n"
            "r_symndex:  %08x\n"
            "r_type:     %1x\n"
            "r_extern:   %1x\n"
            "r_offset:   %1x\n"
            "r_size:     %1x\n",
            r->r_vaddr,
            r->r_symndx,
            r->r_type,
            r->r_extern,
            r->r_offset,
            r->r_size);
    
    fprintf(fp,"r_symndx:   %s\n",
            r->r_symndx <= MAX_R_SN ? sn[r->r_symndx] : "SYMBOL");

    fprintf(fp,"r_type:     %s\n",
            r->r_type < MAX_R_TYPE ? type[r->r_type] : "ILLEGAL");
}
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintRELOC2"

static void PrintRELOC2(FILE *fp,RELOC *r)
{
    
    fprintf(fp,"%016lx %s %s %s\n",
            r->r_vaddr,
            r->r_type < MAX_R_TYPE ? type[r->r_type] : "ILLEGAL",
            r->r_extern ? "EXT" : "INT",
            r->r_symndx <= MAX_R_SN ? sn[r->r_symndx] : "SYMBOL");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SysMarkRelocAndSymbol"

extern void SysMarkRelocAndSymbol(FILE *fp)
{
    INDEX isec;
    INDEX is;
    INDEX iins;
        
#if 1
    FOREACH_SEC(isec)
    {
        if( !AsecIsCode(isec) ) continue; 
        iins = Addr2Ins( ASEC_START(isec) );
        AINS_FLAGS(iins) |= FI_FUN_BEG | FI_BBL_BEG | F_GPREL32;
    }
#endif    
    
    FOREACH_ESY(is)  /* do external symbols */
    {
        if( AESY_IS_PROC(is) && AESY_VALUE(is) != 0 )
        {
            iins = Addr2Ins(AESY_VALUE(is));
            AINS_FLAGS(iins) |= FI_FUN_BEG | FI_BBL_BEG;
            if(iins)
                AINS_FLAGS(iins-1) |= FI_BBL_END | FI_FUN_END;
        }
        

    }

    FOREACH_LSY(is) /* do local symbols */
    {
        if( ALSY_IS_PROC(is) && ALSY_VALUE(is) != 0)
        {
            iins = Addr2Ins(ALSY_VALUE(is));
            AINS_FLAGS(iins) |= FI_FUN_BEG | FI_BBL_BEG;
            if(iins)
                AINS_FLAGS(iins-1) |= FI_BBL_END | FI_FUN_END;
        }
        

    }
    
    FOREACH_SEC(isec)    /* Do the Relocation Info */
    {
        INDEX irel;
        INDEX iins;
        ADDRESS address;
        static ADDRESS immed;
        
        FOREACH_SEC_REL(isec,irel)
        {
            if( ASEC_AREL_EXTERN(isec,irel) )
                continue;

            if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
                ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
                ASEC_AREL_SECTION(isec,irel) != R_SN_NULL &&
                ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
                continue;

            /*
              Check the different reloc types. currently we handle only 3
             */

            switch( ASEC_AREL_TYPE(isec,irel) )
            {
            case R_HINT:
            case R_LITUSE:
            case R_BRADDR:
            case R_GPDISP:
                CASSERT(AsecIsCode(isec),("strange reloc in non code sec\n"));
                break;


                /* I do not know what to do with those, seems to be fine to ignore them - rm*/
                
            case R_ABS:
            case R_OP_PUSH:
            case R_OP_PSUB:
                break;
                
            
            case R_REFQUAD:
                CASSERT(!AsecIsCode(isec),("refquad in code sec\n"));
                
                address = ReadData64(ASEC_AREL_ADDRESS(isec,irel));
                iins = Addr2Ins(address);
                AINS_FLAGS(iins) |= F_REFQUAD;
                AINS_FLAGS(iins) |= FI_BBL_BEG;
                if( iins > 0 )
                    AINS_FLAGS(iins-1) |= FI_BBL_END;
                
                if( IsFunctionStart(address) & !(AINS_FLAGS(iins) & FI_FUN_BEG) )
                {
#if 0
                    WARNING(("REFQUAD for non function at %lx (%s)\n",
                             address,FindSymbol(address)));
#endif           
                    AINS_FLAGS(iins) |=  FI_FUN_BEG;
                    if( iins > 0 )
                        AINS_FLAGS(iins-1) |= FI_FUN_END;
                }
                break;

            case R_GPREL32:

                /* NEEDS WORK ??? */
                CASSERT(!AsecIsCode(isec),("gprel32 in code sec\n"));
                iins = Addr2Ins(BIN_GP+ReadData32(ASEC_AREL_ADDRESS(isec,irel)));
                AINS_FLAGS(iins) |= F_GPREL32 | FI_BBL_BEG;
                if( iins )
                    AINS_FLAGS(iins-1) |= FI_BBL_END;

                break;

            case R_IMMED:

                if( ASEC_AREL_SIZE(isec,irel) == R_IMMED_SCN_HI32 )
                {
                    INDEX iins = Addr2Ins(ASEC_AREL_ADDRESS(isec,irel));
                    CASSERT( AINS_CODE(iins) == I_ldah,("weird\n"));
                    immed =  (long)AINS_DATA(iins)<<16L;
                }

                if( ASEC_AREL_SIZE(isec,irel) == R_IMMED_LO32 )
                {
                    INDEX iins = Addr2Ins(ASEC_AREL_ADDRESS(isec,irel));
                    INDEX target;
                    CASSERT( AINS_CODE(iins) == I_lda,("weird\n"));
                    immed +=  (long)AINS_DATA(iins);
                    immed += ASEC_START(isec);

                    /*
                      This relocation type typically occurs with
                      jsr in the fini or init section
                      we reduce those to bsr here rather than letting the
                      partial evaluator do the job! 
                      */

                    WARNING(("IMMED   for function at %lx (%s) (REFQUAD)\n",immed,FindSymbol(immed)));
                    CASSERT( AINS_REG_C(iins) == REG_PV,("unexpected IMMED reloc reg\n"));
                    iins++;
                    CASSERT( AINS_CODE(iins) == I_jsr,("unexpected IMMED reloc ins\n"));
                    target = Addr2Ins(immed);
                    AINS_CODE(iins) = I_bsr;
                    AINS_REG_B(iins) = REG_ZERO;
                    AINS_DATA(iins) = target - (iins+1);

                    AINS_FLAGS(target) |= F_REFQUAD;
                    AINS_FLAGS(target) |= FI_BBL_BEG;
                    if( iins > 0 )
                        AINS_FLAGS(target-1) |= FI_BBL_END;
                    AINS_FLAGS(target) |=  FI_FUN_BEG;
                    if( iins > 0 )
                        AINS_FLAGS(target-1) |= FI_FUN_END;
                }
                break;
                
            default:
                ERROR(("unexpected reloc\n"));
                PrintRELOC2(fp,&ASEC_AREL(isec,irel));
            }
        }
    }

    InfoStatus(fp,"\t[MarkRelocAndSymbol]\n");
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "SysRelocate"

extern void SysRelocate(FILE *fp)
{
    INDEX isec;
    INDEX irel;
    ADDRESS address;
    
    FOREACH_SEC(isec)    
    {
        FOREACH_SEC_REL(isec,irel)
        {
            if( ASEC_AREL_EXTERN(isec,irel) )continue;

            if( ASEC_AREL_SECTION(isec,irel) != R_SN_TEXT &&
                ASEC_AREL_SECTION(isec,irel) != R_SN_INIT &&
		ASEC_AREL_SECTION(isec,irel) != R_SN_NULL && 
                ASEC_AREL_SECTION(isec,irel) != R_SN_FINI )
                continue;
            
            /* we ignore IMMED reloctypes here */
            /* recompute REFQUAD and GPREL32 relocation info to text segment entries */

            if( ASEC_AREL_TYPE(isec,irel) ==  R_REFQUAD )
            {
                address = ReadData64(ASEC_AREL_ADDRESS(isec,irel));
#if 0
                address = GetNewAddress(address,pair);
#else
                address = TranslateAddress(address);
#endif                
                WriteData64(ASEC_AREL_ADDRESS(isec,irel),address);
            }

            if( ASEC_AREL_TYPE(isec,irel) ==  R_GPREL32 )
            {
                address = BIN_GP + ReadData32(ASEC_AREL_ADDRESS(isec,irel));
#if 0
                address = GetNewAddress(address,pair);
#else
                address = TranslateAddress(address);
#endif                
                WriteData32(ASEC_AREL_ADDRESS(isec,irel),(INT32)(address-BIN_GP));

                /* NEEDS WORK */ 
            }
        }
    }

}

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintSCNHDR"

static void PrintSCNHDR(FILE *fp, SCNHDR *s)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(s,("nil SCNHDR pointer\n"));

    fprintf(fp,
            "s_name:     %.8s\n"
            "s_paddr:    %016lx\n"
            "s_vaddr:    %016lx\n"
            "s_size:     %016lx\n"
            "s_scnptr:   %016lx\n"
            "s_relptr:   %016lx\n"
            "s_lnnoptr:  %016lx\n"
            "s_nreloc:   %04hx\n"
            "s_nlnno:    %04hx\n"
            "s_flags:    %08x\n",
            s->s_name,
            s->s_paddr,
            s->s_vaddr,
            s->s_size,
            s->s_scnptr,
            s->s_relptr,
            s->s_lnnoptr,
            s->s_nreloc,
            s->s_nlnno,
            s->s_flags);

    fprintf(fp,"s_flags:    ");

    if( IS_STYP_REG(s->s_flags) )      fprintf(fp,"STYP_REG ");
    if( IS_STYP_DSECT(s->s_flags) )    fprintf(fp,"STYP_DSECT ");
    if( IS_STYP_NOLOAD(s->s_flags) )   fprintf(fp,"STYP_NOLOAD ");
    if( IS_STYP_GROUP(s->s_flags) )    fprintf(fp,"STYP_GROUP ");
    if( IS_STYP_PAD(s->s_flags) )      fprintf(fp,"STYP_PAD ");
    if( IS_STYP_COPY(s->s_flags) )     fprintf(fp,"STYP_COPY ");
    if( IS_STYP_TEXT(s->s_flags) )     fprintf(fp,"STYP_TEXT ");
    if( IS_STYP_DATA(s->s_flags) )     fprintf(fp,"STYP_DATA ");
    if( IS_STYP_BSS(s->s_flags) )      fprintf(fp,"STYP_BSS ");
    if( IS_STYP_RDATA(s->s_flags) )    fprintf(fp,"STYP_RDATA ");
    if( IS_STYP_SDATA(s->s_flags) )    fprintf(fp,"STYP_SDATA ");
    if( IS_STYP_SBSS(s->s_flags) )     fprintf(fp,"STYP_SBSS ");
    if( IS_STYP_UCODE(s->s_flags) )    fprintf(fp,"STYP_UCODE ");
    if( IS_STYP_GOT(s->s_flags) )      fprintf(fp,"STYP_GOT ");
    if( IS_STYP_DYNAMIC(s->s_flags) )  fprintf(fp,"STYP_DYNAMIC ");
    if( IS_STYP_DYNSYM(s->s_flags) )   fprintf(fp,"STYP_DYNSYM ");
    if( IS_STYP_REL_DYN(s->s_flags) )  fprintf(fp,"STYP_REL_DYN ");
    if( IS_STYP_DYNSTR(s->s_flags) )   fprintf(fp,"STYP_DYNSTR ");
    if( IS_STYP_HASH(s->s_flags) )     fprintf(fp,"STYP_HASH ");
    if( IS_STYP_CONFLICT(s->s_flags) ) fprintf(fp,"STYP_CONFLICT ");
    if( IS_STYP_REGINFO(s->s_flags) )  fprintf(fp,"STYP_REGINFO ");
    if( IS_STYP_FINI(s->s_flags) )     fprintf(fp,"STYP_FINI ");
    if( IS_STYP_INIT(s->s_flags) )     fprintf(fp,"STYP_INIT ");
    if( IS_STYP_COMMENT(s->s_flags) )   fprintf(fp,"STYP_COMMENT ");
    if( IS_STYP_RCONST(s->s_flags) )   fprintf(fp,"STYP_RCONST ");
    if( IS_STYP_XDATA(s->s_flags) )    fprintf(fp,"STYP_XDATA ");
    if( IS_STYP_PDATA(s->s_flags) )    fprintf(fp,"STYP_PDATA ");
    if( IS_STYP_LITA(s->s_flags) )     fprintf(fp,"STYP_LITA ");
    if( IS_STYP_LIT8(s->s_flags) )     fprintf(fp,"STYP_LIT8 ");
    if( IS_STYP_LIT4(s->s_flags) )     fprintf(fp,"STYP_LIT4 ");
    if( IS_S_NRELOC_OVFL(s->s_flags) ) fprintf(fp,"S_NRELOC_OVFL ");
    fprintf(fp,"\n");
}
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintSec"

extern void PrintSec(FILE *fp, INDEX isec)
{
    PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintSym"

static char *stTab[] =
{
    "stNil",
    "stGlobal",
    "stStatic",
    "stParam",
    "stLocal",
    "stLabel",
    "stProc",
    "stBlock",
    "stEnd",
    "stMember",
    "stTypedef",
    "stFile",
    "stStaticProc",
    "stConstant",
    "stStaParam",
    "stBase",
    "stVirtBase",
    "stTag",
    "stInter"
};

static char *scTab[] =
{
"scText",
"scData",
"scBss",
"scRegister",
"scAbs",
"scUndefined",
"scCdbLocal",
"scBits",
"scCdbSystem",
"scDbx",
"scRegImage",
"scInfo",
"scUserStruct",
"scSData",
"scSBss",
"scRData",
"scVar",
"scCommon",
"scSCommon",
"scVarRegister",
"scVariant",
"scFileDesc",
"scSUndefined",
"scInit",
"scBasedVar",
"scReportDesc",
"scXData",
"scPData",
"scFini",
"scRConst"
};
            
/*=======================================================
=======================================================*/

static void PrintSym(FILE *fp, SYMR *s, STRING name)
{
    fprintf(fp,"VAL: %016lx  NM: %08x (%-20s) TY: %1x (%-10s) CL: %1x (%-10s)  AX: %08x\n",
            s->value,
            s->iss,
            name,
            s->st,
            s->st < 21 ? stTab[s->st] : "ILLEGAL",  /* stMax is too big */
            s->sc,
            s->sc < scMax ? scTab[s->sc] : "ILLEGAL",
            s->index);
}
            
/*=======================================================
=======================================================*/

extern void PrintEsy(FILE *fp, INDEX iesy)
{
    PrintSym(fp,&AESY(iesy),AESY_NAME(iesy));
}
            
/*=======================================================
=======================================================*/

extern void PrintLsy(FILE *fp, INDEX ilsy)
{
    PrintSym(fp,&ALSY(ilsy),ALSY_NAME(ilsy));
}
            
/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "PrintPdt"

extern void PrintPdt(FILE *fp,INDEX ip)
{
    PDR *p = &APDR(ip);
                   
    fprintf(fp,
            "adr:        %016lx\n"
            "cbLineOff.: %016lx\n"
            "isym:       %08x (%s)\n"
            "iline:      %08x\n"
            "regmask:    %08x\n"
            "regoffset:  %08x\n"
            "iopt:       %08x\n"
            "fregmask:   %08x\n"
            "fregoffset: %08x\n"
            "frameoff.:  %08x\n"
            "lnLow:      %08x\n"
            "lnHigh:     %08x\n"
            "gp_prologue: %1x\n"
            "gp_used:    %1x\n"
            "reg_frame:  %1x\n"
            "localoff:   %1x\n"
            "framereg:   %1x\n"
            "reg_frame:  %04hx\n"
            "pcreg:      %04hx\n",
            p->adr,
            p->cbLineOffset,
            p->isym,
            ASYM_NAME(p->isym),
            p->iline,
            p->regmask,
            p->regoffset,
            p->iopt,
            p->fregmask,
            p->fregoffset,
            p->frameoffset,
            p->lnLow,
            p->lnHigh,
            p->gp_prologue,
            p->gp_used,
            p->reg_frame,
            p->localoff,
            p->framereg,
            p->reg_frame,
            p->pcreg);
}

            
#endif            
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintHDRR"

static void PrintHDDR(FILE *fp,HDRR *h)
{
    fprintf(fp,
            "magic:      %04hx\n"
            "vstamp:     %04hx\n"
            "ilineMax:   %08x\n"
            "idnMax:     %08x\n"
            "ipdMax:     %08x\n"
            "isymMax:    %08x\n"
            "ioptMax:    %08x\n"
            "iauxMax:    %08x\n"
            "issMax:     %08x\n"
            "issExtMax:  %08x\n"
            "ifdMax:     %08x\n"
            "crfd:       %08x\n"
            "iextMax:    %08x\n"
            "cbLine:     %016lx\n"
            "cbLineOff.: %016lx\n"
            "cbDnOff.:   %016lx\n"
            "cbPdOff.:   %016lx\n"
            "cbSymOff.:  %016lx\n"
            "cbOptOff.t: %016lx\n"
            "cbAuxOff.:  %016lx\n"
            "cbSsOff.:   %016lx\n"
            "cbFdOff.:   %016lx\n"
            "cbRfdOff.:  %016lx\n"
            "cbExtOff.:  %016lx\n",
            h->magic,
            h->vstamp,
            h->ilineMax,
            h->idnMax,
            h->ipdMax,
            h->isymMax,
            h->ioptMax,
            h->iauxMax,
            h->issMax,
            h->issExtMax,
            h->ifdMax,
            h->crfd,
            h->iextMax,
            h->cbLine,
            h->cbLineOffset,
            h->cbDnOffset,
            h->cbPdOffset,
            h->cbSymOffset,
            h->cbOptOffset,
            h->cbAuxOffset,
            h->cbSsOffset,
            h->cbFdOffset,
            h->cbRfdOffset,
            h->cbExtOffset);
}

    
/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintAOUTHDR"

static void PrintAOUTHDR(FILE *fp, AOUTHDR *a)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(a,("nil AOUTHDR pointer\n"));
    
    fprintf(fp,
            "magic:      %04hx\n"
            "vstamp:     %04hx\n"
            "bldrev:     %04hx\n"
            "tsize:      %016lx\n"
            "dsize:      %016lx\n"
            "bsize:      %016lx\n"
            "entry:      %016lx\n"
            "text_start: %016lx\n"
            "data_start: %016lx\n"
            "bss_start:  %016lx\n"
            "gprmask:    %08x\n"
            "fprmask:    %08x\n"
            "gp_value:   %016lx\n",
            a->magic,
            a->vstamp,
            a->bldrev,
            a->tsize,
            a->dsize,
            a->bsize,
            a->entry,
            a->text_start,
            a->data_start,
            a->bss_start,
            a->gprmask,
            a->fprmask,
            a->gp_value);
}


/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "PrintFILEHDR"

static void PrintFILEHDR(FILE *fp,FILHDR *f)
{
    CASSERT(fp,("nil FILE pointer\n"));
    CASSERT(f,("nil FILHDR pointer\n"));
    
    fprintf(fp,
            "f_magic:    %04hx\n"
            "f_nscns:    %04hx\n"
            "f_timdat:   %08x\n"
            "f_symptr:   %016lx\n"
            "f_nsyms:    %08x\n"
            "f_opthdr:   %04hx\n"
            "f_flags:    %04hx\n",
            f->f_magic,
            f->f_nscns,
            f->f_timdat,
            f->f_symptr,
            f->f_nsyms,
            f->f_opthdr,
            f->f_flags);
    
    fprintf(fp,"f_flags:    ");
    if( f->f_flags & F_RELFLG ) fprintf(fp,"F_RELFLG ");
    if( f->f_flags & F_EXEC )   fprintf(fp,"F_EXEC ");
    if( f->f_flags & F_LNNO )   fprintf(fp,"F_LNNO ");
    if( f->f_flags & F_LSYMS )  fprintf(fp,"F_LSYMS ");
    if( f->f_flags & F_NO_SHARED ) fprintf(fp,"F_NO_SHARED ");
    if( f->f_flags & F_LOMAP ) fprintf(fp,"F_LOMAP ");
    fprintf(fp,"\n");
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllHead"

extern void PrintAllHead(FILE *fp)
{
    INDEX isec;

    PrintFILEHDR(fp,&BIN_FILHDR);
    PrintSep2(fp);
    PrintAOUTHDR(fp,&BIN_AOUTHDR);
    PrintSep2(fp);

    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
        PrintSep2(fp);
    }

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllData"

extern void PrintAllData(FILE *fp)
{
    INDEX isec;

    for(isec=0; isec<BIN_NSECTIONS; isec++)
    {
        if( ASEC_IS_TEXT(isec) || ASEC_IS_BSS(isec) || ASEC_IS_SBSS (isec)) continue;

        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));
        PrintHexDump(fp,ASEC_START(isec),(UINT8*)ASEC_DATA8(isec),ASEC_SIZE(isec));
    }

}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllCode"

extern void PrintAllCode(FILE *fp)
{
    INDEX isec;

    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        INDEX iins;
                
        if( !ASEC_IS_TEXT(isec) ) continue;

        FOREACH_INS(iins)
        {
            PrintIns(fp,iins);
        }
    }
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllReloc"

extern void PrintAllReloc(FILE *fp)
{
    INDEX isec;

    for(isec=0;isec<BIN_NSECTIONS;isec++)
    {
        INDEX irel;
        PrintSCNHDR(fp,&ASEC_SCNHDR(isec));

        for( irel=ASEC_HAS_RELOC_OVFL(isec) ? 1 : 0; irel<ASEC_NRELOCS(isec); irel++ )
        {
            PrintRELOC2(fp,&ASEC_AREL(isec,irel));
        }
    }

}

/*=======================================================
=======================================================*/
#if 0
#undef __PROC__
#define __PROC__ "PrintAllFdr"

void PrintAllFdr(FILE *fp)
{
    INDEX ifdr;


    FOREACH_FDR(ifdr)
    {
        PrintFdr(fp,ifdr);
        PrintNL(fp);
    }

}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllPdr"

void PrintAllPdr(FILE *fp)
{
    INDEX ipdr;

    FOREACH_PDR(ipdr)
    {
        PrintPdt(fp,ipdr);
        PrintNL(fp);
    }

}
#endif
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "PrintAllSym"

extern void PrintAllSym(FILE *fp)
{
    INDEX isym;

    fprintf(fp,"\nLOCALS:\n\n");
    
    FOREACH_LSY(isym)
    {
        PrintLsy(fp,isym);
    }

    fprintf(fp,"\nEXTERNALS:\n\n");
        
    FOREACH_ESY(isym)
    {
        PrintEsy(fp,isym);
    }
}


